from __future__ import annotations
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal as TypingLiteral, Union
from lark.indenter import PythonIndenter
from lark import Lark, Transformer, v_args
from lark.tree import Meta

fpy_grammar_str = (Path(__file__).parent / "grammar.lark").read_text()

input_text = None


def parse(text: str):
    parser = Lark(
        fpy_grammar_str,
        start="input",
        parser="lalr",
        postlex=PythonIndenter(),
        propagate_positions=True,
        maybe_placeholders=True,
    )

    global input_text
    input_text = text
    tree = parser.parse(text, on_error=lambda x: print("Error"))
    transformed = FpyTransformer().transform(tree)
    return transformed


@dataclass
class Ast:
    meta: Meta = field(repr=False)
    id: int = field(init=False, repr=False, default=None)
    node_text: str = field(init=False, repr=False, default=None)

    def __post_init__(self):
        if not hasattr(self.meta, "start_pos"):
            self.node_text = ""
            return
        self.node_text = (
            input_text[self.meta.start_pos : self.meta.end_pos]
            .replace("\n", " ")
            .strip()
        )

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.node_text})"


@dataclass
class AstVar(Ast):
    var: str


@dataclass()
class AstString(Ast):
    value: str


@dataclass
class AstNumber(Ast):
    value: int | float


@dataclass
class AstBoolean(Ast):
    value: TypingLiteral[True] | TypingLiteral[False]


AstLiteral = AstString | AstNumber | AstBoolean


@dataclass
class AstGetAttr(Ast):
    parent: "AstReference"
    attr: str


@dataclass
class AstGetItem(Ast):
    parent: "AstReference"
    item: AstNumber


@dataclass
class AstFuncCall(Ast):
    func: "AstReference"
    args: list["AstExpr"] | None


@dataclass
class AstInfixOp(Ast):
    value: str


@dataclass()
class AstPass(Ast):
    pass


@dataclass
class AstComparison(Ast):
    lhs: "AstExpr"
    op: AstInfixOp
    rhs: "AstExpr"


@dataclass
class AstNot(Ast):
    value: "AstExpr"


@dataclass
class AstAnd(Ast):
    values: list["AstExpr"]


@dataclass
class AstOr(Ast):
    values: list["AstExpr"]


AstTest = AstOr | AstAnd | AstNot | AstComparison


AstReference = AstGetAttr | AstGetItem | AstVar
AstExpr = Union[AstFuncCall, AstTest, AstLiteral, AstReference]


@dataclass
class AstAssign(Ast):
    variable: AstVar
    var_type: AstReference | None
    value: AstExpr


@dataclass
class AstElif(Ast):
    condition: AstExpr
    body: "AstBody"


@dataclass
class AstElifs(Ast):
    cases: list[AstElif]


@dataclass()
class AstIf(Ast):
    condition: AstExpr
    body: "AstBody"
    elifs: AstElifs | None
    els: Union["AstBody", None]


AstStmt = Union[AstExpr, AstAssign, AstPass, AstIf]


@dataclass
class AstBody(Ast):
    stmts: list[AstStmt]


for cls in Ast.__subclasses__():
    cls.__hash__ = Ast.__hash__
    # cls.__repr__ = Ast.__repr__


@v_args(meta=False, inline=False)
def as_list(self, tree):
    return list(tree)


def no_inline_or_meta(type):
    @v_args(meta=False, inline=False)
    def wrapper(self, tree):
        return type(tree)

    return wrapper


def no_inline(type):
    @v_args(meta=True, inline=False)
    def wrapper(self, meta, tree):
        return type(meta, tree)

    return wrapper


def no_meta(type):
    @v_args(meta=False, inline=True)
    def wrapper(self, tree):
        return type(tree)

    return wrapper

def handle_str(meta, s: str):
    return s.strip("'").strip('"')


@v_args(meta=True, inline=True)
class FpyTransformer(Transformer):
    input = no_inline(AstBody)
    pass_stmt = AstPass

    assign = AstAssign

    if_stmt = AstIf
    elifs = no_inline(AstElifs)
    elif_ = AstElif
    body = no_inline(AstBody)
    or_test = no_inline(AstOr)
    and_test = no_inline(AstAnd)
    not_test = AstNot
    comparison = AstComparison
    comp_op = AstInfixOp

    func_call = AstFuncCall
    arguments = no_inline_or_meta(list)

    string = AstString
    number = AstNumber
    boolean = AstBoolean
    name = no_meta(str)
    get_attr = AstGetAttr
    get_item = AstGetItem
    var = AstVar

    NAME = str
    DEC_NUMBER = int
    FLOAT_NUMBER = float
    COMPARISON_OP = str
    STRING = handle_str
    CONST_TRUE = lambda a, b: True
    CONST_FALSE = lambda a, b: False



"""F Prime Framer/Deframer Implementation of the CCSDS Space Packet Protocol"""

from __future__ import annotations

import struct
import copy

from spacepackets.ccsds.spacepacket import SpacePacketHeader, PacketType, SpacePacket

from fprime_gds.common.communication.framing import FramerDeframer
from fprime_gds.plugin.definitions import gds_plugin_implementation, gds_plugin
from fprime_gds.common.utils.data_desc_type import DataDescType

from .apid import APID
import logging

LOGGER = logging.getLogger("framing")


@gds_plugin(FramerDeframer)
class SpacePacketFramerDeframer(FramerDeframer):
    """Concrete implementation of FramerDeframer supporting SpacePacket protocol

    This implementation is registered as a "framing" plugin to support encryption within the GDS layer.
    """

    SEQUENCE_COUNT_MAXIMUM = 16384  # 2^14
    HEADER_SIZE = 6
    IDLE_APID = 0x7FF  # max 11 bit value per protocol specification

    def __init__(self):
        # self.sequence_number = 0
        # Map APID to sequence counts
        self.apid_to_sequence_count_map = dict()
        for key in DataDescType:
            self.apid_to_sequence_count_map[key.value] = 0

    def frame(self, data):
        """Frame the supplied data in Space Packet"""
        # The protocol defines length token to be number of bytes minus 1
        data_length_token = len(data) - 1
        apid = APID.from_data(data)
        space_header = SpacePacketHeader(
            packet_type=PacketType.TC,
            apid=apid,
            seq_count=self.get_sequence_count(apid),
            data_len=data_length_token,
        )
        space_packet = SpacePacket(space_header, sec_header=None, user_data=data)
        return space_packet.pack()

    def deframe(self, data, no_copy=False):
        """Deframe the supplied data according to Space Packet protocol"""
        discarded = b""
        if data is None:
            return None, None, discarded
        if not no_copy:
            data = copy.copy(data)
        # Deframe all packets until there is not enough data for a header
        while len(data) >= self.HEADER_SIZE:
            # Read header information including start token and size and check if we have enough for the total size
            try:
                sp_header = SpacePacketHeader.unpack(data)
            except ValueError:
                # If the header is invalid, rotate away a byte and keep processing
                discarded += data[0:1]
                data = data[1:]
                continue
            if sp_header.ccsds_version != 0 or sp_header.packet_type != PacketType.TM:
                # Space Packet version is specified as 0 per protocol
                discarded += data[0:1]
                data = data[1:]
                continue
            # Skip Idle Packets as they are not meaningful
            if sp_header.apid == self.IDLE_APID:
                data = data[sp_header.packet_len :]
                continue
            # Check sequence count and warn if not expected value (don't drop the packet)
            if sp_header.seq_count != self.get_sequence_count(sp_header.apid):
                LOGGER.warning(
                    f"APID {sp_header.apid} received sequence count: {sp_header.seq_count}"
                    f" (expected: {self.get_sequence_count(sp_header.apid)})"
                )
                # Set the sequence count to the next expected value (consider missing packets have been lost)
                self.apid_to_sequence_count_map[sp_header.apid] = (
                    sp_header.seq_count + 1
                )
            # If the pool is large enough to read the whole packet, then read it
            if len(data) >= sp_header.packet_len:
                deframed = struct.unpack_from(
                    # data_len is number of bytes minus 1 per SpacePacket spec
                    f">{sp_header.data_len + 1}s",
                    data,
                    self.HEADER_SIZE,
                )[0]
                data = data[sp_header.packet_len :]
                LOGGER.debug(f"Deframed packet: {sp_header}")
                return deframed, data, discarded
            else:
                # If we don't have enough data, then break out of the loop
                break
        return None, data, discarded

    def get_sequence_count(self, apid: int):
        """Get the sequence number and increment

        This function will return the current sequence number and then increment the sequence number for the next round.
        Should an APID not be registered already, it will be initialized to 0.

        Return:
            current sequence number
        """
        # If APID is not registered, initialize it to 0
        sequence = self.apid_to_sequence_count_map.get(apid, 0)
        self.apid_to_sequence_count_map[apid] = (
            sequence + 1
        ) % self.SEQUENCE_COUNT_MAXIMUM
        return sequence

    @classmethod
    def get_name(cls):
        """Name of this implementation provided to CLI"""
        return "raw-space-packet"

    @classmethod
    @gds_plugin_implementation
    def register_framing_plugin(cls):
        """Register the MyPlugin plugin"""
        return cls



"""
standard.py:

This file contains the necessary functions used to setup a standard pipeline that will pull data from the middle-ware
layer into the python system. This layer is designed to setup the base necessary to running Gds objects up-to the layer
of decoding. This enables users to generate add-in that function above the decoders, and run on a standard pipeline
below.

:author: lestarch
"""

import datetime
import os.path
from pathlib import Path
from typing import Type

import fprime.common.models.serialize.time_type

import fprime_gds.common.data_types.cmd_data
import fprime_gds.common.distributor.distributor
import fprime_gds.common.logger.data_logger
from fprime_gds.common.transport import RoutingTag, ThreadedTCPSocketClient
from fprime_gds.common.utils.config_manager import ConfigManager

# Local imports for the sake of composition
from . import dictionaries, encoding, files, histories


class StandardPipeline:
    """
    Class used to encapsulate all of the components of a standard pipeline. The life-cycle of this class follows the
    basic steps:
       1. setup: creates objects needed to read from middle-ware layer and provide data up the stack
       2. register: consumers from this pipeline should register to receive decoder callbacks
       3. run: this pipeline should either take over the standard thread of execution, or be executed on another thread
       4. terminate: called to shutdown the pipeline
    This class provides for basic log files as a fallback for storing events as well. These logs are stored in a given
    directory, which is created on the initialization of this class.
    """

    def __init__(self):
        """
        Set core variables to None or their composition handlers.
        """
        self.distributor = None
        self.client_socket = None
        self.logger = None
        self.dictionary_path = None
        self.up_store = None
        self.down_store = None

        self.__dictionaries = dictionaries.Dictionaries()
        self.__coders = encoding.EncodingDecoding()
        self.__histories = histories.Histories()
        self.__filing = files.Filing()
        self.__transport_type = ThreadedTCPSocketClient

    def setup(
        self,
        config: ConfigManager,
        dictionary,
        file_store,
        logging_prefix=None,
        packet_spec=None,
        packet_set_name=None,
        data_logging_enabled=True
    ):
        """
        Setup the standard pipeline for moving data from the middleware layer through the GDS layers using the standard
        patterns. This allows just registering the consumers, and invoking 'setup' all other of the GDS support layer.

        :param config: config object used when constructing the pipeline.
        :param dictionary: dictionary path. Used to setup loading of dictionaries.
        :param file_store: uplink/downlink storage directory
        :param logging_prefix: logging prefix. Defaults to not logging at all.
        :param packet_spec: location of packetized telemetry XML specification.
        """
        assert (
            dictionary is not None and Path(dictionary).is_file()
        ), f"Dictionary {dictionary} does not exist"
        # File storage configuration for uplink and downlink
        self.up_store = Path(file_store) / "fprime-uplink"
        self.down_store = Path(file_store) / "fprime-downlink"
        try:
            self.down_store.mkdir(parents=True, exist_ok=True)
            self.up_store.mkdir(parents=True, exist_ok=True)
        except PermissionError:
            raise PermissionError(
                f"{file_store} is not writable. Fix permissions or change storage directory with --file-storage-directory."
            )
        self.dictionary_path = Path(dictionary)
        # Loads the distributor and client socket
        self.distributor = fprime_gds.common.distributor.distributor.Distributor(config)
        self.client_socket = self.__transport_type()
        # Setup dictionaries encoders and decoders
        self.dictionaries.load_dictionaries(
            self.dictionary_path, packet_spec, packet_set_name
        )        
        # Update config to use Fw types defined in the JSON dictionary
        if self.dictionaries.fw_type_name:
            for fw_type_name, fw_type in self.dictionaries.fw_type_name.items():
                config.set("types", fw_type_name, fw_type)
        self.coders.setup_coders(
            self.dictionaries, self.distributor, self.client_socket, config
        )
        self.histories.setup_histories(self.coders)
        self.files.setup_file_handling(
            self.down_store,
            self.coders.file_encoder,
            self.coders.file_decoder,
            self.distributor,
            logging_prefix,
        )
        # Register distributor to client socket
        self.client_socket.register(self.distributor)
        # Final setup step is to make a logging directory, and register in the logger
        if logging_prefix and data_logging_enabled:
            self.setup_logging(logging_prefix)

    @property
    def transport_implementation(self):
        """Get implementation type for transport"""
        return self.__transport_type

    @transport_implementation.setter
    def transport_implementation(self, transport_type: Type[None]):
        """Set the implementation type for transport"""
        assert (
            self.client_socket is None
        ), "Cannot setup transport implementation type after setup"
        self.__transport_type = transport_type

    @classmethod
    def get_dated_logging_dir(cls, prefix=os.path.expanduser("~")):
        """
        Sets up the dated subdirectory based upon a given prefix

        :param prefix:
        :return: Path to new directory where logs will be stored for this pipeline
        """
        # Setup log file location
        dts = datetime.datetime.now()
        log_dir = os.path.join(prefix, dts.strftime("%Y-%m-%dT%H:%M:%S.%f"))
        # Make the directories if they do not exit
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        return log_dir

    def setup_logging(self, log_dir):
        """
        Setup logging based on the logging prefix supplied

        :param log_dir: logging output directory
        """
        # Setup the logging pipeline (register it to all its data sources)
        logger = fprime_gds.common.logger.data_logger.DataLogger(
            log_dir, verbose=True, csv=True
        )
        self.logger = logger
        self.coders.register_channel_consumer(self.logger)
        self.coders.register_event_consumer(self.logger)
        self.coders.register_command_consumer(self.logger)
        self.coders.register_packet_consumer(self.logger)
        self.client_socket.register(self.logger)

    def connect(
        self, connection_uri, incoming_tag=RoutingTag.GUI, outgoing_tag=RoutingTag.FSW
    ):
        """Connects to the middleware layer

        Connect to the middleware layer. This connection needs to identify if the object is a a FSW or GUI client. The
        default connection acts as a GUI client sending to FSW.

        Note: this function provides backwards compatibility with historical versions of this connect method of the form
        pipeline.connect(host, port). This version explicitly supplies a hostname without port, and integer port. No
        other arguments are excepted. Use is discouraged.

        Args:
            connection_uri: URI of the connection to make
            incoming_tag: this pipeline will act as supplied tag (GUI, FSW). Default: GUI
            outgoing_tag: this pipeline will produce data for supplied tag (FSW, GUI). Default: FSW
        """
        # Backwards compatibility with the old method .connect(host, port)
        if (
            isinstance(incoming_tag, int)
            and ":" not in connection_uri
            and outgoing_tag == RoutingTag.FSW
        ):
            connection_uri = f"{connection_uri}:{incoming_tag}"
            incoming_tag = RoutingTag.GUI
        self.client_socket.connect(connection_uri, incoming_tag, outgoing_tag)

    def disconnect(self):
        """
        Disconnect from socket
        """
        try:
            if self.client_socket is not None:
                self.client_socket.disconnect()
        finally:
            if self.files is not None and self.files.uplinker is not None:
                self.files.uplinker.exit()

    def send_command(self, command, args):
        """Sends commands to the encoder and history.

        :param command: command id from dictionary to get command template
        :param args: arguments to process
        """
        if isinstance(command, str):
            command_template = self.dictionaries.command_name[command]
        else:
            command_template = self.dictionaries.command_id[command]
        cmd_data = fprime_gds.common.data_types.cmd_data.CmdData(
            tuple(args), command_template
        )
        cmd_data.time = fprime.common.models.serialize.time_type.TimeType()
        cmd_data.time.set_datetime(datetime.datetime.now(), 2)
        self.coders.send_command(cmd_data)

    @property
    def dictionaries(self):
        """
        Get a dictionaries object

        :return: dictionaries composition
        """
        return self.__dictionaries

    @property
    def coders(self):
        """
        Get a coders object

        :return: coders composition
        """
        return self.__coders

    @property
    def histories(self):
        """
        Get a histories object

        :return: histories composition
        """
        return self.__histories

    @property
    def files(self):
        """
        Files member property

        :return: filing compositions
        """
        return self.__filing



"""
@brief Loader class for importing xml based event dictionaries

@date Created July 19, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime_gds.common.data_types import exceptions
from fprime_gds.common.templates.event_template import EventTemplate
from fprime_gds.common.utils.event_severity import EventSeverity

# Custom Python Modules
from .xml_loader import XmlLoader


class EventXmlLoader(XmlLoader):
    """Class to load xml based event dictionaries"""

    EVENT_SECT = "events"

    COMP_TAG = "component"
    NAME_TAG = "name"
    ID_TAG = "id"
    SEVERITY_TAG = "severity"
    FMT_STR_TAG = "format_string"
    DESC_TAG = "description"

    def construct_dicts(self, path):
        """
        Constructs and returns python dictionaries keyed on id and name

        This function should not be called directly, instead, use
        get_id_dict(path) and get_name_dict(path)

        Args:
            path: Path to the xml dictionary file containing event information

        Returns:
            A tuple with two event dictionaries (python type dict):
            (id_idct, name_dict). The keys are the events' id and name fields
            respectively and the values are ChTemplate objects
        """
        xml_tree = self.get_xml_tree(path)
        versions = xml_tree.attrib.get(
            "framework_version", "unknown"
        ), xml_tree.attrib.get("project_version", "unknown")

        # Check if xml dict has events section
        event_section = self.get_xml_section(self.EVENT_SECT, xml_tree)
        if event_section is None:
            msg = f"Xml dict did not have a {self.EVENT_SECT} section"
            raise exceptions.GseControllerParsingException(msg)

        id_dict = {}
        name_dict = {}

        for event in event_section:
            event_dict = event.attrib

            event_comp = event_dict[self.COMP_TAG]
            event_name = event_dict[self.NAME_TAG]
            event_id = int(event_dict[self.ID_TAG], base=16)
            event_severity = EventSeverity[event_dict[self.SEVERITY_TAG]]
            event_fmt_str = XmlLoader.preprocess_format_str(
                event_dict[self.FMT_STR_TAG]
            )

            event_desc = None
            if self.DESC_TAG in event_dict:
                event_desc = event_dict[self.DESC_TAG]

            # Parse arguments
            args = self.get_args_list(
                event, xml_tree, f"{ event_comp }::{ event_name }"
            )

            event_temp = EventTemplate(
                event_id,
                event_name,
                event_comp,
                args,
                event_severity,
                event_fmt_str,
                event_desc,
            )

            id_dict[event_id] = event_temp
            name_dict[event_temp.get_full_name()] = event_temp

        return id_dict, name_dict, versions



"""
event_json_loader.py:

Loads flight dictionary (JSON) and returns id and mnemonic based Python dictionaries of events

@author thomas-bc
"""

from fprime_gds.common.templates.event_template import EventTemplate
from fprime_gds.common.utils.event_severity import EventSeverity
from fprime_gds.common.loaders.json_loader import JsonLoader
from fprime_gds.common.data_types.exceptions import GdsDictionaryParsingException


class EventJsonLoader(JsonLoader):
    """Class to load json based event dictionaries"""

    EVENTS_FIELD = "events"

    NAME = "name"
    ID = "id"
    SEVERITY = "severity"
    FMT_STR = "format"
    DESC = "annotation"
    PARAMETERS = "formalParams"

    def construct_dicts(self, _):
        """
        Constructs and returns python dictionaries keyed on id and name

        This function should not be called directly, instead, use
        get_id_dict(path) and get_name_dict(path)

        Args:
            _: Unused argument (inherited)

        Returns:
            A tuple with two event dictionaries (python type dict):
            (id_dict, name_dict). The keys are the events' id and name fields
            respectively and the values are ChTemplate objects
        """
        id_dict = {}
        name_dict = {}

        if self.EVENTS_FIELD not in self.json_dict:
            raise GdsDictionaryParsingException(
                f"Ground Dictionary missing '{self.EVENTS_FIELD}' field: {str(self.json_file)}"
            )

        for event_dict in self.json_dict[self.EVENTS_FIELD]:
            event_temp = self.construct_template_from_dict(event_dict)

            id_dict[event_temp.get_id()] = event_temp
            name_dict[event_temp.get_full_name()] = event_temp

        return (
            dict(sorted(id_dict.items())),
            dict(sorted(name_dict.items())),
            self.get_versions(),
        )

    def construct_template_from_dict(self, event_dict: dict):
        try:
            qualified_component_name, event_name = event_dict[self.NAME].rsplit('.', 1)
            event_id = event_dict[self.ID]
            event_severity = EventSeverity[event_dict[self.SEVERITY]]
        except ValueError as e:
            raise GdsDictionaryParsingException(
                f"Event dictionary entry malformed, expected name of the form '<QUAL_COMP_NAME>.<EVENT_NAME>' in : {str(event_dict)}"
            )
        except KeyError as e:
            raise GdsDictionaryParsingException(
                f"{str(e)} key missing from Event dictionary entry: {str(event_dict)}"
            )

        event_fmt_str = JsonLoader.preprocess_format_str(
            event_dict.get(self.FMT_STR, "")
        )

        event_desc = event_dict.get(self.DESC)

        # Parse arguments
        event_args = []
        for arg in event_dict.get(self.PARAMETERS, []):
            try:
                arg_name = arg["name"]
                arg_type = self.parse_type(arg["type"])
            except KeyError as e:
                raise GdsDictionaryParsingException(
                    f"{str(e)} key missing from Event parameter or its associated type in the dictionary: {str(arg)}"
                )
            event_args.append(
                (
                    arg_name,
                    arg.get("annotation"),
                    arg_type,
                )
            )

        return EventTemplate(
            event_id,
            event_name,
            qualified_component_name,
            event_args,
            event_severity,
            event_fmt_str,
            event_desc,
        )



import pytest
import struct

from fprime_gds.common.communication.ccsds.space_data_link import (
    SpaceDataLinkFramerDeframer,
)

SCID_TEST_VALUE = 0x77
VCID_TEST_VALUE = 5


@pytest.fixture
def framer_deframer():
    return SpaceDataLinkFramerDeframer(scid=SCID_TEST_VALUE, vcid=VCID_TEST_VALUE)


def test_frame_valid_data(framer_deframer):
    """Test framing valid data."""
    data = b"test_payload"
    framed_data = framer_deframer.frame(data)
    expected_length = (
        len(data)
        + SpaceDataLinkFramerDeframer.TC_TRAILER_SIZE
        + SpaceDataLinkFramerDeframer.TC_HEADER_SIZE
    )
    assert data in framed_data
    assert len(framed_data) == expected_length
    scid = ((framed_data[0] & 0x03) << 2) | framed_data[1]
    assert scid == SCID_TEST_VALUE
    vcid = (framed_data[2] & 0xFC) >> 2
    assert vcid == VCID_TEST_VALUE


def test_deframe_valid_frame(framer_deframer):
    """Test deframing a valid TM frame."""
    FIXED_PAYLOAD_LENGTH = (
        framer_deframer.TM_FIXED_FRAME_SIZE
        - SpaceDataLinkFramerDeframer.TM_HEADER_SIZE
        - SpaceDataLinkFramerDeframer.TM_TRAILER_SIZE
    )
    global_vcid_u16 = (SCID_TEST_VALUE << 4) | (VCID_TEST_VALUE << 1)
    mc_count_u8 = 0
    vc_count_u8 = 0
    status_u16 = 0
    payload = bytes([i % 256 for i in range(FIXED_PAYLOAD_LENGTH)])
    input_data_no_crc = (
        struct.pack(
            ">HBBH",
            global_vcid_u16,
            mc_count_u8,
            vc_count_u8,
            status_u16,
        )
        + payload
    )
    crc = framer_deframer.CRC_CALCULATOR.checksum(input_data_no_crc)
    input_data = input_data_no_crc + struct.pack(">H", crc)
    deframed_data, remaining_data, discarded = framer_deframer.deframe(input_data)
    assert deframed_data == payload
    assert remaining_data == b""
    assert discarded == b""

def test_deframe_incorrect_crc(framer_deframer):
    """Test deframing a valid TM frame."""
    FIXED_PAYLOAD_LENGTH = (
        framer_deframer.TM_FIXED_FRAME_SIZE
        - SpaceDataLinkFramerDeframer.TM_HEADER_SIZE
        - SpaceDataLinkFramerDeframer.TM_TRAILER_SIZE
    )
    global_vcid_u16 = (SCID_TEST_VALUE << 4) | (VCID_TEST_VALUE << 1)
    mc_count_u8 = 0
    vc_count_u8 = 0
    status_u16 = 0
    payload = bytes([i % 256 for i in range(FIXED_PAYLOAD_LENGTH)])
    input_data_no_crc = (
        struct.pack(
            ">HBBH",
            global_vcid_u16,
            mc_count_u8,
            vc_count_u8,
            status_u16,
        )
        + payload
    )
    crc = framer_deframer.CRC_CALCULATOR.checksum(input_data_no_crc) + 1  # Intentionally incorrect CRC
    input_data = input_data_no_crc + struct.pack(">H", crc)
    deframed_data, remaining_data, discarded = framer_deframer.deframe(input_data)
    assert deframed_data is None
    assert remaining_data == input_data[1:]
    assert discarded[0] == input_data[0]



"""
gds_test_api.py:

This file contains basic asserts that can support integration tests on an FPrime
deployment. This API uses the standard pipeline to get access to commands, events,
telemetry and dictionaries.

:author: koran
"""

import signal
import time
from pathlib import Path
import shutil
import json

from fprime.common.models.serialize.time_type import TimeType

from fprime_gds.common.handlers import DataHandler
from fprime_gds.common.history.chrono import ChronologicalHistory
from fprime_gds.common.history.test import TestHistory
from fprime_gds.common.logger.test_logger import TestLogger
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.utils.event_severity import EventSeverity


class IntegrationTestAPI(DataHandler):
    """
    A value used to begin searches after the current contents in a history and only search future
    items
    """

    NOW = "NOW"

    def __init__(self, pipeline, deployment_config=None, logpath=None, fsw_order=True):
        """
        Initializes API: constructs and registers test histories.
        Args:
            pipeline: a pipeline object providing access to basic GDS functionality
            deployment_config: path to deployment configuration file
            logpath: an optional output destination for the api test log
            fsw_order: a flag to determine whether the API histories will maintain FSW time order.
        """
        self.pipeline = pipeline
        self.deployment_config = deployment_config

        # these are owned by the GDS and will not be modified by the test API.
        self.aggregate_command_history = pipeline.histories.commands
        self.aggregate_telemetry_history = pipeline.histories.channels
        self.aggregate_event_history = pipeline.histories.events

        # these histories are owned by the TestAPI and are modified by the API.
        self.fsw_ordered = fsw_order
        if fsw_order:
            self.command_history = ChronologicalHistory()
            self.telemetry_history = ChronologicalHistory()
            self.event_history = ChronologicalHistory()
        else:
            self.command_history = TestHistory()
            self.telemetry_history = TestHistory()
            self.event_history = TestHistory()
        self.pipeline.coders.register_command_consumer(self.command_history)
        self.pipeline.coders.register_event_consumer(self.event_history)
        self.pipeline.coders.register_channel_consumer(self.telemetry_history)

        # Initialize latest time. Will be updated whenever a time query is made.
        self.latest_time = TimeType()

        # Initialize the logger
        self.logger = TestLogger(logpath) if logpath is not None else None

        # Copy dictionaries and binary file to output directory
        if logpath is not None:
            base_dir = Path(self.pipeline.dictionary_path).parents[1]
            for subdir in ["bin", "dict"]:
                dir_path = base_dir / subdir
                if dir_path.is_dir():
                    shutil.copytree(
                        dir_path, Path(logpath) / subdir, dirs_exist_ok=True
                    )

        # A predicate used as a filter to choose which events to log automatically
        self.event_log_filter = self.get_event_pred()

        # Used by the data_callback method to detect if events have been received out of order.
        self.last_evr = None

    def setup(self):
        """Set up the API, assumes pipeline is now setup"""
        self.pipeline.coders.register_event_consumer(self)

    def teardown(self):
        """
        To be called once at the end of the API's use. Closes the test log and clears histories.
        """
        self.clear_histories()
        if self.logger is not None:
            self.logger.close_log()
            self.logger = None

    ######################################################################################
    #   API Functions
    ######################################################################################
    def start_test_case(self, case_name, case_id):
        """
        To be called at the start of a test case. This function inserts a log message to denote a
        new test case is beginning, records the latest time stamp in case the user clears the
        aggregate histories, and then clears the API's histories.

        Args:
            case_name: the name of the test case (str)
            case_id: a short identifier to denote the test case (str or number)
        """
        msg = f"[STARTING CASE] {case_name}"
        self.__log(msg, TestLogger.GRAY, TestLogger.BOLD, case_id=case_id)
        self.get_latest_time()  # called in case aggregate histories are cleared by the user
        self.clear_histories()

    def log(self, msg, color=None):
        """
        User-accessible function to log user messages to the test log.
        Args:
            msg: a user-provided message to add to the test log. (str)
            color: a string containing a color hex code "######" (str)
        """
        self.__log(msg, color, sender="API user")

    def get_latest_time(self):
        """
        Finds the latest flight software time received by either history.

        Returns:
            a flight software timestamp (TimeType)
        """
        events = self.aggregate_event_history.retrieve()
        e_time = TimeType()
        if len(events) > 0:
            e_time = events[-1].get_time()

        channels = self.aggregate_telemetry_history.retrieve()
        t_time = TimeType()
        if len(channels) > 0:
            t_time = channels[-1].get_time()

        self.latest_time = max(e_time, t_time, self.latest_time)
        return self.latest_time

    def test_assert(self, value, msg="", expect=False):
        """
        this assert gives the user the ability to add formatted assert messages to the test log and
        raise an assertion.
        Args:
            value: a boolean value that determines if the assert is successful.
            msg: a string describing what is checked by the assert.
            expect: when True, the call will behave as an expect, and will skip the assert (boolean)
        Return:
            True if the assert was successful, False otherwise
        """
        if not expect:
            ast_msg = "User assertion"
            fail_color = TestLogger.RED
        else:
            ast_msg = "User expectation"
            fail_color = TestLogger.ORANGE

        if value:
            ast_msg = f"{ast_msg} succeeded: {msg}"
            self.__log(ast_msg, TestLogger.GREEN)
        else:
            ast_msg = f"{ast_msg} failed: {msg}"
            self.__log(ast_msg, fail_color)

        if not expect:
            assert value, ast_msg
        return value

    def predicate_assert(self, predicate, value, msg="", expect=False):
        """
        API assert gives the user the ability to add formatted assert messages to the test log and
        raise an assertion.
        Args:
            value: the value to be evaluated by the predicate. (object)
            predicate: an instance of predicate that will decided if the test passes (predicate)
            msg: a string describing what is checked by the assert. (str)
            expect: when True, the call will behave as an expect, and will skip the assert (boolean)
        Return:
            True if the assert was successful, False otherwise
        """
        return self.__assert_pred("User", predicate, value, msg, expect)

    def clear_histories(self, time_stamp=None):
        """
        Clears the IntegrationTestAPI's histories. Because the command history is not correlated to
        a flight software timestamp, it will be cleared entirely. This function can be used to set
        up test cases so that the IntegrationTestAPI's histories only contain objects received
        during that test.
        Note: this will not clear user-created sub-histories nor the aggregate histories (histories
        owned by the GDS)

        Args:
            time_stamp: If specified, histories are only cleared before the timestamp.
        """
        if time_stamp is not None:
            time_pred = predicates.greater_than_or_equal_to(time_stamp)
            e_pred = predicates.event_predicate(time_pred=time_pred)
            self.event_history.clear(e_pred)
            t_pred = predicates.telemetry_predicate(time_pred=time_pred)
            self.telemetry_history.clear(t_pred)
            msg = f"Clearing Test Histories after {time_stamp}"
        else:
            self.event_history.clear()
            self.telemetry_history.clear()
            msg = "Clearing Test Histories"

        self.__log(msg, TestLogger.WHITE)
        self.command_history.clear()

    def set_event_log_filter(
        self, event=None, args=None, severity=None, time_pred=None
    ):
        """
        Constructs an event predicate that is then used to filter which events are interlaced in the
        test logs. This method replaces the current filter. Calling this method with no arguments
        will effectively reset the filter.

        Args:
            event: an optional mnemonic (str), id (int), or predicate to specify the event type
            args: an optional list of arguments (list of values, predicates, or None to ignore)
            severity: an EventSeverity enum or a predicate to specify the event severity
            time_pred: an optional predicate to specify the flight software timestamp
        """
        self.event_log_filter = self.get_event_pred(event, args, severity, time_pred)

    def get_deployment(self):
        """
        Get the deployment of the target using the loaded FSW dictionary.

        Returns:
            The name of the deployment (str) or None if not found
        """
        dictionary = str(self.pipeline.dictionary_path)

        try:
            with open(dictionary, "r") as file:
                data = json.load(file)
            return data["metadata"].get("deploymentName")
        except FileNotFoundError:
            msg = f"Error: File not found at path: {dictionary}"
            self.__log(msg, TestLogger.YELLOW)
            return None
        except json.JSONDecodeError as e:
            msg = f"Error decoding JSON: {e}"
            self.__log(msg, TestLogger.YELLOW)
            return None
        except Exception as e:
            msg = f"An unexpected error occurred: {e} is an unknown key"
            self.__log(msg, TestLogger.YELLOW)
            return None

    def wait_for_dataflow(self, count=1, channels=None, start=None, timeout=120):
        """
        Wait for data flow by checking for any telemetry updates within a specified timeout.

        Args:
            count: either an exact amount (int) or a predicate to specify how many objects to find
            channels: a channel specifier or list of channel specifiers (mnemonic, ID, or predicate). All will count if None
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        """
        if start is None:
            start = self.get_latest_time()

        history = self.get_telemetry_subhistory()
        result = self.await_telemetry_count(
            count, channels=channels, history=history, start=start, timeout=timeout
        )
        if not result:
            msg = f"Failed to detect any data flow for {timeout} s."
            self.__log(msg, TestLogger.RED)
            assert False, msg
        self.remove_telemetry_subhistory(history)

    def get_config_file_path(self):
        """
        Accessor for IntegrationTestAPI's deployment configuration file.

        Returns:
            path to user-specified deployment configuration file (str) or None if not defined
        """
        if self.deployment_config:
            return self.deployment_config
        else:
            return None

    def load_config_file(self):
        """
        Load user-specified deployment configuration JSON file.

        Returns:
            JSON object as a dictionary
        """
        config_file = self.get_config_file_path()

        try:
            with open(config_file, "r") as file:
                result = json.load(file)
            return result
        except FileNotFoundError:
            msg = f"Error: File not found at path {config_file}"
            self.__log(msg, TestLogger.RED)
            assert False, msg
        except json.JSONDecodeError as e:
            msg = f"Error decoding JSON: {e}"
            self.__log(msg, TestLogger.RED)
            assert False, msg
        except Exception as e:
            msg = f"An unexpected error occurred: {e}"
            self.__log(msg, TestLogger.RED)
            assert False, msg

    def get_mnemonic(self, comp=None, name=None):
        """
        Get deployment mnemonic of specified item from user-specified deployment
        configuration file.

        Args:
            comp: qualified name of the component instance (str), i.e. "<component>.<instance>"
            name: command, channel, or event name (str) [optional]
        Returns:
            deployment mnemonic of specified item (str) or native mnemonic (str) if not found
        """
        data = self.load_config_file()

        if data:
            try:
                mnemonic = data[comp]
                return f"{mnemonic}.{name}" if name else f"{mnemonic}"
            except KeyError:
                self.__log(f"Error: {comp} not found", TestLogger.YELLOW)
                return f"{comp}.{name}" if name else f"{comp}"
        else:
            return f"{comp}.{name}" if name else f"{comp}"

    def get_prm_db_path(self) -> str:
        """
        Get file path to parameter db from user-specified deployment configuration file.

        Returns:
            file path to parameter db (str) or None if not found
        """
        data = self.load_config_file()

        if data:
            try:
                filepath = data["Svc.PrmDb.filename"]
                if filepath.startswith("/"):
                    return filepath
                else:
                    msg = f"Error: {filepath} did not start with a forward slash"
                    self.__log(msg, TestLogger.RED)
                    assert False, msg
            except KeyError:
                return None
        else:
            return None

    ######################################################################################
    #   History Functions
    ######################################################################################
    def get_command_test_history(self):
        """
        Accessor for IntegrationTestAPI's command history
        Returns:
            a history of CmdData objects
        """
        return self.command_history

    def get_telemetry_test_history(self):
        """
        Accessor for IntegrationTestAPI's telemetry history
        Returns:
            a history of ChData objects
        """
        return self.telemetry_history

    def get_event_test_history(self):
        """
        Accessor for IntegrationTestAPI's event history
        Returns:
            a history of EventData objects
        """
        return self.event_history

    def get_telemetry_subhistory(self, telemetry_filter=None, fsw_order=True):
        """
        Returns a new instance of TestHistory that will be updated with new telemetry updates as
        they come in. Specifying a filter will only enqueue updates that satisfy the filter in this
        new sub-history. The returned history can be substituted into the await and assert methods
        of this API.

        Args:
            telemetry_filter: an optional predicate used to filter a subhistory.
            fsw_order: a flag to determine whether this subhistory will maintain FSW time order.
        Returns:
            an instance of TestHistory
        """
        if fsw_order:
            subhist = ChronologicalHistory(telemetry_filter)
        else:
            subhist = TestHistory(telemetry_filter)
        self.pipeline.coders.register_channel_consumer(subhist)
        return subhist

    def remove_telemetry_subhistory(self, subhist):
        """
        De-registers the subhistory from the GDS. Once called, the given subhistory will stop
        receiving telemetry updates.

        Args:
            subhist: a TestHistory instance that is subscribed to event messages
        Returns:
            True if the subhistory was removed, False otherwise
        """
        return self.pipeline.coders.remove_channel_consumer(subhist)

    def get_event_subhistory(self, event_filter=None, fsw_order=True):
        """
        Returns a new instance of TestHistory that will be updated with new events as they come in.
        Specifying a filter will only enqueue events that satisfy the filter in this new
        sub-history. The returned history can be substituted into the await and assert methods of
        this API.

        Args:
            event_filter: an optional predicate to filter a subhistory.
            fsw_order: a flag to determine whether this subhistory will maintain FSW time order.
        Returns:
            an instance of TestHistory
        """
        if fsw_order:
            subhist = ChronologicalHistory(event_filter)
        else:
            subhist = TestHistory(event_filter)
        self.pipeline.coders.register_event_consumer(subhist)
        return subhist

    def remove_event_subhistory(self, subhist):
        """
        De-registers the subhistory from the GDS. Once called, the given subhistory will stop
        receiving event messages.

        Args:
            subhist: a TestHistory instance that is subscribed to event messages
        Returns:
            True if the subhistory was removed, False otherwise
        """
        return self.pipeline.coders.remove_event_consumer(subhist)

    ######################################################################################
    #   Command Functions
    ######################################################################################
    def translate_command_name(self, command):
        """
        This function will translate the given mnemonic into an ID as defined by the flight
        software dictionary. This call will raise an error if the command given is not in the
        dictionary.

        Args:
            command: Either the command id (int) or the command mnemonic (str)

        Returns:
            The command ID (int)
        """
        if isinstance(command, str):
            cmd_dict = self.pipeline.dictionaries.command_name
            if command in cmd_dict:
                return cmd_dict[command].get_id()
            msg = f"The command mnemonic, {command}, wasn't in the dictionary"
        else:
            cmd_dict = self.pipeline.dictionaries.command_id
            if command in cmd_dict:
                return command
            msg = f"The command id, {command}, wasn't in the dictionary"
        raise KeyError(msg)

    def send_command(self, command, args=None):
        """
        Sends the specified command.
        Args:
            command: the mnemonic (str) or ID (int) of the command to send
            args: a list of command arguments.
        """
        if args is None:
            args = []

        msg = f"Sending Command: {command} {args}"
        self.__log(msg, TestLogger.PURPLE)
        command = self.translate_command_name(command)
        self.pipeline.send_command(command, args)

    def send_and_await_telemetry(self, command, args=None, channels=None, timeout=5):
        """
        Sends the specified command and awaits the specified channel update or sequence of
        updates. See await_telemetry and await_telemetry_sequence for full details.
        Note: If awaiting a sequence avoid specifying timestamps.

        Args:
            command: the mnemonic (str) or ID (int) of the command to send
            args: a list of command arguments.
            channels: a single or a sequence of channel specs (event_predicates, mnemonics, or IDs)
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)

        Returns:
            The channel update or updates found by the search
        """
        if args is None:
            args = []
        if channels is None:
            channels = []

        start = self.telemetry_history.size()
        self.send_command(command, args)
        if isinstance(channels, list):
            return self.await_telemetry_sequence(channels, start=start, timeout=timeout)
        return self.await_telemetry(channels, start=start, timeout=timeout)

    def send_and_await_event(self, command, args=None, events=None, timeout=5):
        """
        Sends the specified command and awaits the specified event message or sequence of
        messages. See await_event and await event sequence for full details.
        Note: If awaiting a sequence avoid specifying timestamps.

        Args:
            command: the mnemonic (str) or ID (int) of the command to send
            args: a list of command arguments.
            events: a single or a sequence of event specifiers (event_predicates, mnemonics, or IDs)
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)

        Returns:
            The event or events found by the search
        """
        if args is None:
            args = []
        if events is None:
            events = []

        start = self.event_history.size()
        self.send_command(command, args)
        if isinstance(events, list):
            return self.await_event_sequence(events, start=start, timeout=timeout)
        return self.await_event(events, start=start, timeout=timeout)

    def send_and_assert_command(
        self,
        command,
        args=[],
        max_delay=None,
        timeout=5,
        events=None,
        commander="cmdDisp",
    ):
        """
        This helper will send a command and verify that the command was dispatched and completed
        within the F' deployment. This helper can retroactively check that the delay between
        dispatch and completion is less than a maximum allowable delay.

        Args:
            command: the mnemonic (str) or ID (int) of the command to send
            args: a list of command arguments.
            max_delay: the maximum allowable delay between dispatch and completion (int/float)
            timeout: the number of seconds to wait before terminating the search (int)
            events: extra event predicates to check between  dispatch and complete
            commander: the command dispatching component. Defaults to cmdDisp
        Return:
            returns a list of the EventData objects found by the search
        """
        cmd_id = self.translate_command_name(command)
        dispatch = [
            self.get_event_pred(f"{commander}.OpCodeDispatched", [cmd_id, None])
        ]
        complete = [self.get_event_pred(f"{commander}.OpCodeCompleted", [cmd_id])]
        events = dispatch + (events if events else []) + complete
        results = self.send_and_assert_event(command, args, events, timeout=timeout)
        if max_delay is not None:
            delay = results[1].get_time() - results[0].get_time()
            msg = f"The delay, {delay}, between the two events should be < {max_delay}"
            assert delay < max_delay, msg
        return results

    ######################################################################################
    #   Command Asserts
    ######################################################################################
    def send_and_assert_telemetry(self, command, args=None, channels=None, timeout=5):
        """
        Sends the specified command and asserts on the specified channel update or sequence of
        updates. See await_telemetry and await_telemetry_sequence for full details.
        Note: If awaiting a sequence avoid specifying timestamps.

        Args:
            command: the mnemonic (str) or ID (int) of the command to send
            args: a list of command arguments.
            channels: a single or a sequence of channel specs (event_predicates, mnemonics, or IDs)
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)

        Returns:
            The channel update or updates found by the search
        """
        if args is None:
            args = []
        if channels is None:
            channels = []

        start = self.telemetry_history.size()
        self.send_command(command, args)
        if isinstance(channels, list):
            return self.assert_telemetry_sequence(
                channels, start=start, timeout=timeout
            )
        return self.assert_telemetry(channels, start=start, timeout=timeout)

    def send_and_assert_event(self, command, args=None, events=None, timeout=5):
        """
        Sends the specified command and asserts on the specified event message or sequence of
        messages. See assert_event and assert event sequence for full details.

        Args:
            command: the mnemonic (str) or ID (int) of the command to send
            args: a list of command arguments.
            events: a single or a sequence of event specifiers (event_predicates, mnemonics, or IDs)
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)

        Returns:
            The event or events found by the search
        """
        if args is None:
            args = []
        if events is None:
            events = []

        start = self.event_history.size()
        self.send_command(command, args)
        if isinstance(events, list):
            return self.assert_event_sequence(events, start=start, timeout=timeout)
        return self.assert_event(events, start=start, timeout=timeout)

    ######################################################################################
    #   Telemetry Functions
    ######################################################################################
    def translate_telemetry_name(self, channel, force_component=True):
        """
        This function will translate the given mnemonic into an ID as defined by the flight
        software dictionary. This call will raise an error if the channel given is not in the
        dictionary.

        Args:
            channel: a channel mnemonic (str) or id (int)
            force_component: (optional) force the event to match a fully qualified event name when supplying a string
        Returns:
            the channel ID (int) or list of channel ids when force component is False
        """
        if isinstance(channel, str):
            ch_dict = self.pipeline.dictionaries.channel_name
            matching = [
                ch_dict[name].get_id()
                for name in ch_dict.keys()
                if name.endswith(f".{channel}")
            ]
            if channel in ch_dict:
                return ch_dict[channel].get_id()
            if force_component or not matching:
                msg = f"The telemetry mnemonic, {channel}, wasn't in the dictionary"
                raise KeyError(msg)
            return matching

        ch_dict = self.pipeline.dictionaries.channel_id
        if channel in ch_dict:
            return channel
        msg = f"The telemetry mnemonic, {channel}, wasn't in the dictionary"
        raise KeyError(msg)

    def get_telemetry_pred(self, channel=None, value=None, time_pred=None):
        """
        This function will translate the channel ID, and construct a telemetry_predicate object. It
        is used as a helper by the IntegrationTestAPI, but could also be helpful to a user of the
        test API. If  channel is already an instance of telemetry_predicate, it will be returned
        immediately. The provided implementation of telemetry_predicate evaluates true if and only
        if all specified constraints are satisfied. If a specific constraint isn't specified, then
        it will not effect the outcome; this means all arguments are optional. If no constraints
        are specified, the predicate will always return true.


        Args:
            channel: an optional mnemonic (str), id (int), or predicate to specify the channel type
            value: an optional value (object/number) or predicate to specify the value field
            time_pred: an optional predicate to specify the flight software timestamp
        Returns:
            an instance of telemetry_predicate
        """
        if isinstance(channel, predicates.telemetry_predicate):
            return channel

        if not predicates.is_predicate(channel) and channel is not None:
            channel = self.translate_telemetry_name(channel, force_component=False)
            channel = (
                predicates.is_a_member_of(channel)
                if isinstance(channel, list)
                else predicates.equal_to(channel)
            )

        if not predicates.is_predicate(value) and value is not None:
            value = predicates.equal_to(value)

        return predicates.telemetry_predicate(channel, value, time_pred)

    def await_telemetry(
        self, channel, value=None, time_pred=None, history=None, start="NOW", timeout=5
    ):
        """
        A search for a single telemetry update received. By default, the call will only await
        until a correct update is found. The user can specify that await also searches the current
        history by specifying a value for start. On timeout, the search will return None.

        Args:
            channel: a channel specifier (mnemonic, id, or predicate)
            value: optional value (object/number) or predicate to specify the value field
            time_pred: an optional predicate to specify the flight software timestamp
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            the ChData object found during the search, otherwise, None
        """
        t_pred = self.get_telemetry_pred(channel, value, time_pred)

        if history is None:
            history = self.get_telemetry_test_history()

        return self.find_history_item(t_pred, history, start, timeout)

    def await_telemetry_sequence(self, channels, history=None, start="NOW", timeout=5):
        """
        A search for a sequence of telemetry updates. By default, the call will only await until
        the sequence is completed. The user can specify that await also searches the history by
        specifying a value for start. On timeout, the search will return the list of found
        channel updates regardless of whether the sequence is complete.
        Note: It is recommended (but not enforced) not to specify timestamps for this assert.
        Note: This function will always return a list of updates. The user should check if the
        sequence was completed.

        Args:
            channels: an ordered list of channel specifiers (mnemonic, id, or predicate)
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            an ordered list of ChData objects that satisfies the sequence
        """
        seq_preds = [self.get_telemetry_pred(channel) for channel in channels]

        if history is None:
            history = self.get_telemetry_test_history()

        return self.find_history_sequence(seq_preds, history, start, timeout)

    def await_telemetry_count(
        self, count, channels=None, history=None, start="NOW", timeout=5
    ):
        """
        A search on the number of telemetry updates received. By default, the call will only await
        until a correct count is achieved. The user can specify that await also searches the current
        history by specifying a value for start. On timeout, the search will return the list of
        found channel updates regardless of whether a correct count is achieved.
        Note: this search will always return a list of objects. The user should check if the search
        was completed.

        Args:
            count: either an exact amount (int) or a predicate to specify how many objects to find
            channels: a channel specifier or list of channel specifiers (mnemonic, ID, or predicate)
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            a list of the ChData objects that were counted
        """
        if channels is None:
            search = None
        elif isinstance(channels, list):
            t_preds = [self.get_telemetry_pred(channel=channel) for channel in channels]
            search = predicates.satisfies_any(t_preds)
        else:
            search = self.get_telemetry_pred(channel=channels)

        if history is None:
            history = self.get_telemetry_test_history()

        return self.find_history_count(count, history, search, start, timeout)

    ######################################################################################
    #   Telemetry Asserts
    ######################################################################################
    def assert_telemetry(
        self, channel, value=None, time_pred=None, history=None, start=None, timeout=0
    ):
        """
        An assert on a single telemetry update received. If the history doesn't have the
        correct update, the call will await until a correct update is received or the
        timeout, at which point it will assert failure.

        Args:
            channel: a channel specifier (mnemonic, id, or predicate)
            value: optional value (object/number) or predicate to specify the value field
            time_pred: an optional predicate to specify the flight software timestamp
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            the ChData object found during the search
        """
        pred = self.get_telemetry_pred(channel, value, time_pred)
        result = self.await_telemetry(
            channel, value, time_pred, history, start, timeout
        )
        msg = "checks if item search found a correct update"
        self.__assert_pred("Telemetry received", pred, result, msg)
        return result

    def assert_telemetry_sequence(self, channels, history=None, start=None, timeout=0):
        """
        A search for a sing sequence of telemetry updates messages. If the history doesn't have the
        complete sequence, the call will await until the sequence is completed or the timeout, at
        which point it will return the list of found channel updates.
        Note: It is recommended (but not enforced) not to specify timestamps for this assert.
        Note: This function will always return a list of updates the user should check if the
        sequence was completed.

        Args:
            channels: an ordered list of channel specifiers (mnemonic, id, or predicate)
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            an ordered list of ChData objects that satisfies the sequence
        """
        results = self.await_telemetry_sequence(channels, history, start, timeout)
        len_pred = predicates.equal_to(len(channels))
        msg = "checks if sequence search found every update"
        self.__assert_pred("Telemetry sequence", len_pred, len(results), msg)
        return results

    def assert_telemetry_count(
        self, count, channels=None, history=None, start=None, timeout=0
    ):
        """
        An assert on the number of channel updates received. If the history doesn't have the
        correct update count, the call will await until a correct count is achieved or the
        timeout, at which point it will assert failure.

        Args:
            count: either an exact amount (int) or a predicate to specify how many objects to find
            channels: a channel specifier or list of channel specifiers (mnemonic, ID, or predicate)
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            a list of the ChData objects that were counted
        """
        results = self.await_telemetry_count(count, channels, history, start, timeout)
        count_pred = (
            count if predicates.is_predicate(count) else predicates.equal_to(count)
        )
        msg = "checks if count search found a correct amount of updates"
        self.__assert_pred("Telemetry count", count_pred, len(results), msg)
        return results

    ######################################################################################
    #   Event Functions
    ######################################################################################
    def translate_event_name(self, event, force_component=True):
        """
        This function will translate the given mnemonic into an ID as defined by the
        flight software dictionary. This call will raise an error if the event given is
        not in the dictionary.

        Args:
            event: an event mnemonic (str) or ID (int)
            force_component: (optional) force the event to match a fully qualified event name when supplying a string
        Returns:
            the event ID (int) or list of event ids when force component is False
        """
        if isinstance(event, str):
            event_dict = self.pipeline.dictionaries.event_name
            matching = [
                event_dict[name].get_id()
                for name in event_dict.keys()
                if name.endswith(f".{event}")
            ]
            if event in event_dict:
                return event_dict[event].get_id()
            if force_component or not matching:
                msg = f"The event mnemonic, {event}, wasn't in the dictionary"
                raise KeyError(msg)
            return matching

        event_dict = self.pipeline.dictionaries.event_id
        if event in event_dict:
            return event
        msg = f"The event id, {event}, wasn't in the dictionary"
        raise KeyError(msg)

    def get_event_pred(self, event=None, args=None, severity=None, time_pred=None):
        """
        This function will translate the event ID, and construct an event_predicate object. It is
        used as a helper by the IntegrationTestAPI, but could also be helpful to a user of the test
        API. If event is already an instance of event_predicate, it will be returned immediately.
        The provided implementation of event_predicate evaluates true if and only if all specified
        constraints are satisfied. If a specific constraint isn't specified, then it will not
        effect the outcome; this means all arguments are optional. If no constraints are specified,
        the predicate will always return true.

        Args:
            event: mnemonic (str), id (int), or predicate to specify the event type
            args: list of arguments (list of values, predicates, or None to ignore)
            severity: an EventSeverity enum or a predicate to specify the event severity
            time_pred: predicate to specify the flight software timestamp
        Returns:
            an instance of event_predicate
        """
        if isinstance(event, predicates.event_predicate):
            return event

        if not predicates.is_predicate(event) and event is not None:
            event = self.translate_event_name(event, force_component=False)
            event = (
                predicates.is_a_member_of(event)
                if isinstance(event, list)
                else predicates.equal_to(event)
            )

        if not predicates.is_predicate(args) and args is not None:
            args = predicates.args_predicate(args)

        if not predicates.is_predicate(severity) and severity is not None:
            if not isinstance(severity, EventSeverity):
                msg = f"Given severity was not a valid Severity Enum Value: {severity} ({type(severity)})"
                raise TypeError(msg)
            severity = predicates.equal_to(severity)

        return predicates.event_predicate(event, args, severity, time_pred)

    def await_event(
        self,
        event,
        args=None,
        severity=None,
        time_pred=None,
        history=None,
        start="NOW",
        timeout=5,
    ):
        """
        A search for a single event message received. By default, the call will only await until a
        correct message is found. The user can specify that await also searches the current history
        by specifying a value for start. On timeout, the search will return None.

        Args:
            event: an event specifier (mnemonic, id, or predicate)
            args: a list of expected arguments (list of values, predicates, or None for don't care)
            severity: an EventSeverity enum or a predicate to specify the event severity
            time_pred: an optional predicate to specify the flight software timestamp
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            the EventData object found during the search, otherwise, None
        """
        e_pred = self.get_event_pred(event, args, severity, time_pred)

        if history is None:
            history = self.get_event_test_history()

        return self.find_history_item(e_pred, history, start, timeout)

    def await_event_sequence(self, events, history=None, start="NOW", timeout=5):
        """
        A search for a sequence of event messages. By default, the call will only await until
        the sequence is completed. The user can specify that await also searches the history by
        specifying a value for start. On timeout, the search will return the list of found
        event messages regardless of whether the sequence is complete.
        Note: It is recommended (but not enforced) not to specify timestamps for this assert.
        Note: This function will always return a list of events the user should check if the
        sequence was completed.

        Args:
            events: an ordered list of event specifiers (mnemonic, id, or predicate)
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            an ordered list of EventData objects that satisfies the sequence
        """
        seq_preds = [self.get_event_pred(event) for event in events]

        if history is None:
            history = self.get_event_test_history()

        return self.find_history_sequence(seq_preds, history, start, timeout)

    def await_event_count(
        self, count, events=None, history=None, start="NOW", timeout=5
    ):
        """
        A search on the number of events received. By default, the call will only await until a
        correct count is achieved. The user can specify that this await also searches the current
        history by specifying a value for start. On timeout, the search will return the list of
        found event messages regardless of whether a correct count is achieved.
        Note: this search will always return a list of objects. The user should check if the search
        was completed.

        Args:
            count: either an exact amount (int) or a predicate to specify how many objects to find
            events: an event specifier or list of event specifiers (mnemonic, ID, or predicate)
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            a list of the EventData objects that were counted
        """
        if events is None:
            search = None
        elif isinstance(events, list):
            e_preds = [self.get_event_pred(event=event) for event in events]
            search = predicates.satisfies_any(e_preds)
        else:
            search = self.get_event_pred(event=events)

        if history is None:
            history = self.get_event_test_history()

        return self.find_history_count(count, history, search, start, timeout)

    ######################################################################################
    #   Event Asserts
    ######################################################################################
    def assert_event(
        self,
        event,
        args=None,
        severity=None,
        time_pred=None,
        history=None,
        start=None,
        timeout=0,
    ):
        """
        An assert on a single event message received. If the history doesn't have the
        correct message, the call will await until a correct message is received or the
        timeout, at which point it will assert failure.

        Args:
            event: an event specifier (mnemonic, id, or predicate)
            args: a list of expected arguments (list of values, predicates, or None for don't care)
            severity: an EventSeverity enum or a predicate to specify the event severity
            time_pred: an optional predicate to specify the flight software timestamp
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            the EventData object found during the search
        """
        pred = self.get_event_pred(event, args, severity, time_pred)
        result = self.await_event(
            event, args, severity, time_pred, history, start, timeout
        )
        msg = "checks if item search found a correct event"
        self.__assert_pred("Event received", pred, result, msg)
        return result

    def assert_event_sequence(self, events, history=None, start=None, timeout=0):
        """
        An assert that a sequence of event messages is received. If the history doesn't have the
        complete sequence, the call will await until the sequence is completed or the timeout, at
        which point it will assert failure.
        Note: It is recommended (but not enforced) not to specify timestamps for this assert.

        Args:
            events: an ordered list of event specifiers (mnemonic, id, or predicate)
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            an ordered list of EventData objects that satisfied the sequence
        """
        results = self.await_event_sequence(events, history, start, timeout)
        len_pred = predicates.equal_to(len(events))
        msg = "checks if sequence search found every event"
        self.__assert_pred("Event sequence", len_pred, len(results), msg)
        return results

    def assert_event_count(
        self, count, events=None, history=None, start=None, timeout=0
    ):
        """
        An assert on the number of events received. If the history doesn't have the
        correct event count, the call will await until a correct count is achieved or the
        timeout, at which point it will assert failure.

        Args:
            count: either an exact amount (int) or a predicate to specify how many objects to find
            events: optional event specifier or list of specifiers (mnemonic, id, or predicate)
            history: if given, a substitute history that the function will search and await
            start: an optional index or predicate to specify the earliest item to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            a list of the EventData objects that were counted
        """
        results = self.await_event_count(count, events, history, start, timeout)
        count_pred = (
            count if predicates.is_predicate(count) else predicates.equal_to(count)
        )
        msg = "checks if count search found a correct amount of events"
        self.__assert_pred("Event count", count_pred, len(results), msg)
        return results

    ######################################################################################
    #   File Uplink functions
    ######################################################################################

    def uplink_file_and_await_completion(self, file_path, destination=None, timeout=10):
        """
        This function will upload a file and wait for its completion, awaiting for the
        FileReceived event.

        Args:
            file_path: the path to the file to upload
            destination: the destination path for the uploaded file
            timeout: the maximum time to wait for the event
        """
        self.uplink_file(file_path, destination)
        self.await_event("FileReceived", timeout=timeout)

    def uplink_file(self, file_path, destination=None):
        """
        This function will upload a file to the specified location.

        Note: this will simply put the file on the outgoing queue. No guarantee
        is made on when the file will be delivered. To wait for the completion of
        the file uplink, use uplink_file_and_await_completion()

        Args:
            file_path: the path to the file to upload
        """
        uplink_file = Path(self.pipeline.up_store) / Path(file_path).name
        shutil.copy2(file_path, uplink_file)
        self.pipeline.files.uplinker.enqueue(str(uplink_file), destination)

    ######################################################################################
    #   History Searches
    ######################################################################################
    class __HistorySearcher:
        """
        This class defines the calls made by the __history_search helper method and has a unique
        implementation for each type of search provided by the api.
        """

        def __init__(self):
            self.ret_val = None
            self.repeats = False

        def search_current_history(self, items):
            """
            Searches the scoped existing history
            Return:
                True if the search was satisfied, False otherwise
            """
            raise NotImplementedError()

        def incremental_search(self, item):
            """
            Searches one awaited item at a time
            Return:
                True if the search was satisfied, False otherwise
            """
            raise NotImplementedError()

        def get_return_value(self):
            """
            Returns the result of the search whether the search is successful or not
            """
            return self.ret_val

        def requires_repeats(self):
            """
            Returns a flag to determine if the history searcher needs repeated data objects when receive order does not match chronological order.
            """
            return self.repeats

    class TimeoutException(Exception):
        """
        This exception is used by the history searches to signal the end of the timeout.
        """

    def __timeout_sig_handler(self, signum, frame):
        raise self.TimeoutException()

    def __search_test_history(self, searcher, name, history, start=None, timeout=0):
        """
        This helper method contains the common logic to all search methods in the test API. This
        means searches on both the event and channel histories rely on this helper. Each history
        search is performed on both current history items and then on items that have yet to be
        added to the history. The API defines these two scopes using the variables start and
        timeout. They have several useful behaviors.

        start is used to pick the earliest item to search in the current history. start can be
        specified as either a predicate to search for the first item, an index of the history, the
        API variable NOW, or an instance of the TimeType timestamp object. The behavior of NOW is
        to ignore the current history and only search awaited items until the timestamp. The
        behavior of giving a timetype is to only search items that happened at or after the
        specified timestamp.

        timeout is a specification of how long to await future items in seconds. Specifying a
        timeout of 0 will ignore all future items. The timeout specifies an increment of time
        relative to the local clock, not the embedded application's clock.
        Note: the API does not try to check for edge cases where the final item in a search is
        received as the search times out. The user should ensure that their timeouts are sufficient
        to complete any awaiting searches.

        Finally, the test API supports the ability to substitute a history object for any search.
        This is part of why history must be specified for each search

        Args:
            searcher: an instance of __HistorySearcher to execute search-specific logic
            name: a string name to differentiate the type of search
            history: the TestHistory object to conduct the search on
            start: an index, a predicate, the NOW variable, or a TimeType timestamp to pick the
                first item to search
            timeout: the number of seconds to await future items
        """
        if start == self.NOW:
            start = history.size()
        elif isinstance(start, TimeType):
            time_pred = predicates.greater_than_or_equal_to(start)
            e_pred = self.get_telemetry_pred(time_pred=time_pred)
            t_pred = self.get_event_pred(time_pred=time_pred)
            start = predicates.satisfies_any([e_pred, t_pred])

        current = history.retrieve(start)
        if searcher.search_current_history(current):
            return searcher.get_return_value()

        if timeout:
            self.__log(f"{name} now awaiting for at most {timeout} s.")
            check_repeats = isinstance(history, ChronologicalHistory)
            try:
                signal.signal(signal.SIGALRM, self.__timeout_sig_handler)
                signal.alarm(timeout)
                while True:
                    if check_repeats:
                        new_items = history.retrieve_new(searcher.requires_repeats())
                    else:
                        new_items = history.retrieve_new()
                    for item in new_items:
                        if searcher.incremental_search(item):
                            return searcher.get_return_value()
                    time.sleep(0.1)
            except self.TimeoutException:
                self.__log(
                    f"{name} timed out and ended unsuccessfully.", TestLogger.YELLOW
                )
            finally:
                signal.alarm(0)
        else:
            self.__log(f"{name} ended unsuccessfully.", TestLogger.YELLOW)
        return searcher.get_return_value()

    def find_history_item(self, search_pred, history, start=None, timeout=0):
        """
        This function can both search and await for an element in a history. The function will
        return the first valid object it finds. The search will return when an object is found, or
        the timeout is reached.

        Args:
            search_pred: a predicate to specify a history item.
            history: the history that the function will search and await
            start: an index or predicate to specify the earliest item from the history to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            the data object found during the search, otherwise, None
        """

        class __ItemSearcher(self.__HistorySearcher):
            def __init__(self, log, search_pred):
                super().__init__()
                self.log = log
                self.search_pred = search_pred
                self.repeats = False
                self.ret_val = None
                msg = f"Beginning an item search for an item that satisfies:\n    {self.search_pred}"
                self.log(msg, TestLogger.YELLOW)

            def search_current_history(self, items):
                return any(self.incremental_search(item) for item in items)

            def incremental_search(self, item):
                if self.search_pred(item):
                    msg = f"History search found the specified item: {item}"
                    self.log(msg, TestLogger.YELLOW)
                    self.ret_val = item
                    return True
                return False

        searcher = __ItemSearcher(self.__log, search_pred)
        return self.__search_test_history(
            searcher, "Item search", history, start, timeout
        )

    def find_history_sequence(self, seq_preds, history, start=None, timeout=0):
        """
        This function can both search and await for a sequence of elements in a history. The
        function will return a list of the history objects to satisfy the sequence search. The
        search will return when an order of data objects is found that satisfies the entire
        sequence, or the timeout occurs.
        Note: this search will always return a list of objects. The user should check if the search
        was completed.

        Args:
            seq_preds: an ordered list of predicate objects to specify a sequence
            history: the history that the function will search and await
            start: an index or predicate to specify the earliest item from the history to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            a list of data objects that satisfied the sequence
        """

        class __SequenceSearcher(self.__HistorySearcher):
            def __init__(self, log, seq_preds):
                super().__init__()
                self.log = log
                self.ret_val = []
                self.seq_preds = seq_preds.copy()
                self.repeats = True
                msg = f"Beginning a sequence search of {len(self.seq_preds)} items."
                self.log(msg, TestLogger.YELLOW)

            def search_current_history(self, items):
                if len(self.seq_preds) == 0:
                    msg = "Sequence search finished, as the specified sequence had 0 items."
                    self.log(msg, TestLogger.YELLOW)
                    return True

                return any(self.incremental_search(item) for item in items)

            def incremental_search(self, item):
                if self.seq_preds[0](item):
                    self.log(f"Sequence search found the next item: {item}")
                    self.ret_val.append(item)
                    self.seq_preds.pop(0)
                    if len(self.seq_preds) == 0:
                        self.log(
                            "Sequence search found the last item.", TestLogger.YELLOW
                        )
                        return True
                return False

        searcher = __SequenceSearcher(self.__log, seq_preds)
        return self.__search_test_history(
            searcher, "Sequence search", history, start, timeout
        )

    def find_history_count(
        self, count, history, search_pred=None, start=None, timeout=0
    ):
        """
        This function first counts all valid items in the current history, then can await until a
        valid number of elements is received by the history. The function will return a list of the
        history objects counted by the search. The search will return when a correct count of data
        objects is found, or the timeout occurs.
        Note: this search will always return a list of objects. The user should check if the search
        was completed.

        Args:
            count: either an exact amount (int) or a predicate to specify how many objects to find
            history: the history that the function will search and await
            search_pred: a predicate to specify which items to count. If left blank, all will count
            start: an index or predicate to specify the earliest item from the history to search
            timeout: the number of seconds to wait before terminating the search (int)
        Returns:
            a list of data objects that were counted during the search
        """

        class __CountSearcher(self.__HistorySearcher):
            def __init__(self, log, count, search_pred):
                super().__init__()
                self.log = log
                self.ret_val = []
                if predicates.is_predicate(count):
                    self.count_pred = count
                else:
                    self.count_pred = predicates.equal_to(count)
                self.search_pred = search_pred
                self.repeats = False

                msg = f"Beginning a count search for an amount of items ({self.count_pred})."
                self.log(msg, TestLogger.YELLOW)

            def search_current_history(self, items):
                if self.search_pred is None:
                    self.search_pred = predicates.always_true()
                    self.ret_val = items
                else:
                    for item in items:
                        if search_pred(item):
                            self.log(f"Count search counted another item: {item}")
                            self.ret_val.append(item)

                if self.count_pred(len(self.ret_val)):
                    msg = f"Count search found a correct amount: {len(self.ret_val)}"
                    self.log(msg, TestLogger.YELLOW)
                    return True
                return False

            def incremental_search(self, item):
                if self.search_pred(item):
                    self.log(f"Count search counted another item: {item}")
                    self.ret_val.append(item)
                    if self.count_pred(len(self.ret_val)):
                        msg = (
                            f"Count search found a correct amount: {len(self.ret_val)}"
                        )
                        self.log(msg, TestLogger.YELLOW)
                        return True
                return False

        searcher = __CountSearcher(self.__log, count, search_pred)
        return self.__search_test_history(
            searcher, "Count search", history, start, timeout
        )

    ######################################################################################
    #   API Helper Methods
    ######################################################################################
    def __log(self, message, color=None, style=None, sender="Test API", case_id=None):
        """
        Logs and prints an API Message. If the API isn't using a Logger, then only a message will
        be printed.
        Args:
            message: a string containing the message to log
            color: a color hex code (str)
            style: a style option from test logger (str) ["BOLD", "ITALICS", "UNDERLINED"]
            sender: a marker for where the message originated
            case_id: a tag for the current test case. Only needs to be set when a new case starts
        """
        if not isinstance(message, str):
            message = str(message)
        if self.logger:
            self.logger.log_message(message, sender, color, style, case_id)
        # TODO: Re-add way of printing to console?

    def __assert_pred(self, name, predicate, value, msg="", expect=False):
        """
        Helper to assert that a value satisfies a predicate. Includes arguments to provide more
        descriptive messages in the logs.
        Args:
            name: short name describing the check
            predicate: an instance of predicate to determine if the assert is successful
            value: the object evaluated by the predicate
            msg: a string message to describe what the assert is checking
            expect: a boolean value: True will be have as an expect and not raise an assertion.
        Returns:
            True if the assertion was successful, False otherwise
        """
        name = name + (" expectation" if expect else " assertion")
        pred_msg = predicates.get_descriptive_string(value, predicate)
        if predicate(value):
            ast_msg = f"{name} succeeded: {msg}\nassert {pred_msg}"
            self.__log(ast_msg, TestLogger.GREEN)
            if not expect:
                assert True, pred_msg
            return True
        ast_msg = f"{name} failed: {msg}\nassert {pred_msg}"
        if not expect:
            self.__log(ast_msg, TestLogger.RED)
            assert False, pred_msg
        else:
            self.__log(ast_msg, TestLogger.ORANGE)
        return False

    def data_callback(self, data, sender=None):
        """
        Data callback used by the api to log events and detect when they are received out of order.
        Args:
            data: object to store
        """
        if self.event_log_filter(data):
            msg = f"Received EVR: {data.get_str(verbose=True)}"
            self.__log(msg, TestLogger.BLUE, sender="GDS")
        if self.last_evr is not None and data.get_time() < self.last_evr.get_time():
            msg = f"API detected out of order evrs!\nReceived First:{self.last_evr.get_str(verbose=True)}"
            msg += f"\nReceived Second:{data.get_str(verbose=True)}"
            self.__log(msg, TestLogger.ORANGE)
        self.last_evr = data



from dataclasses import astuple, dataclass
from pathlib import Path
import struct
from typing import ClassVar
import zlib
from fprime.common.models.serialize.time_type import TimeType
from fprime.common.models.serialize.numerical_types import (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
    I64Type,
)
from fprime.common.models.serialize.bool_type import BoolType
from enum import Enum

FwSizeType = U64Type
FwChanIdType = U32Type
FwPrmIdType = U32Type
FwOpcodeType = U32Type

MAX_SERIALIZABLE_REGISTER_SIZE = 512 - 4 - 4


class DirectiveOpcode(Enum):
    INVALID = 0
    WAIT_REL = 1
    WAIT_ABS = 2
    SET_SER_REG = 3
    GOTO = 4
    IF = 5
    NO_OP = 6
    GET_TLM = 7
    GET_PRM = 8
    CMD = 9
    SET_REG = 10
    DESER_SER_REG_8 = 11
    DESER_SER_REG_4 = 12
    DESER_SER_REG_2 = 13
    DESER_SER_REG_1 = 14
    # binary reg op directives
    # all of these are handled at the CPP level by one BinaryRegOpDirective
    # boolean ops
    OR = 15
    AND = 16
    # integer equalities
    IEQ = 17
    INE = 18
    # unsigned integer inequalities
    ULT = 19
    ULE = 20
    UGT = 21
    UGE = 22
    # signed integer inequalities
    SLT = 23
    SLE = 24
    SGT = 25
    SGE = 26
    # floating point equalities
    FEQ = 27
    FNE = 28
    # floating point inequalities
    FLT = 29
    FLE = 30
    FGT = 31
    FGE = 32
    # end binary reg op directives

    # unary reg op dirs
    NOT = 33
    # floating point extension and truncation
    FPEXT = 34
    FPTRUNC = 35
    # floating point conversion to signed/unsigned integer,
    # and vice versa
    FPTOSI = 36
    FPTOUI = 37
    SITOFP = 38
    UITOFP = 39
    # end unary reg op dirs

    EXIT = 40


class Directive:
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.INVALID

    def serialize(self) -> bytes:
        arg_bytes = self.serialize_args()

        output = U8Type(self.opcode.value).serialize()
        output += U16Type(len(arg_bytes)).serialize()
        output += arg_bytes

        return output

    def serialize_args(self) -> bytes:
        raise NotImplementedError("serialize_args not implemented")


HEADER_FORMAT = "!BBBBBHI"
HEADER_SIZE = struct.calcsize(HEADER_FORMAT)


@dataclass
class Header:
    majorVersion: int
    minorVersion: int
    patchVersion: int
    schemaVersion: int
    argumentCount: int
    statementCount: int
    bodySize: int


FOOTER_FORMAT = "!I"
FOOTER_SIZE = struct.calcsize(FOOTER_FORMAT)


@dataclass
class Footer:
    crc: int


def serialize_directives(dirs: list[Directive], output: Path = None):
    output_bytes = bytes()

    for dir in dirs:
        output_bytes += dir.serialize()

    header = Header(0, 0, 0, 1, 0, len(dirs), len(output_bytes))
    output_bytes = struct.pack(HEADER_FORMAT, *astuple(header)) + output_bytes

    crc = zlib.crc32(output_bytes) % (1 << 32)
    footer = Footer(crc)
    output_bytes += struct.pack(FOOTER_FORMAT, *astuple(footer))

    if output is None:
        output = input.with_suffix(".bin")

    output.write_bytes(output_bytes)


@dataclass
class WaitRelDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.WAIT_REL
    seconds: int
    useconds: int

    def serialize_args(self) -> bytes:
        return U32Type(self.seconds).serialize() + U32Type(self.useconds).serialize()


@dataclass
class WaitAbsDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.WAIT_ABS
    wakeup_time: TimeType

    def serialize_args(self) -> bytes:
        return self.wakeup_time.serialize()


@dataclass
class SetSerRegDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.SET_SER_REG

    index: int
    """U8: The index of the local variable to set."""
    value: bytes
    """[Fpy.MAX_SERIALIZABLE_REGISTER_SIZE] U8: The value of the local variable."""

    def serialize_args(self) -> bytes:
        data = bytearray()
        data.extend(U8Type(self.index).serialize())
        data.extend(self.value)
        return bytes(data)


@dataclass
class GotoDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.GOTO
    statement_index: int
    """U32: The statement index to execute next."""

    def serialize_args(self) -> bytes:
        return U32Type(self.statement_index).serialize()


@dataclass
class IfDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.IF
    conditional_reg: int
    """U8: The register to branch based off of (interpreted as a C++ boolean)."""
    false_goto_stmt_index: int
    """U32: The statement index to go to if the register is false."""

    def serialize_args(self) -> bytes:
        return (
            U8Type(self.conditional_reg).serialize()
            + U32Type(self.false_goto_stmt_index).serialize()
        )


@dataclass
class NoOpDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.NO_OP

    def serialize_args(self) -> bytes:
        return bytes()


@dataclass
class GetTlmDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.GET_TLM
    value_dest_sreg: int
    """U8: The local variable to store the telemetry value in."""
    time_dest_sreg: int
    """U8: The local variable to store the telemetry time in."""
    chan_id: int
    """FwChanIdType: The telemetry channel ID to get."""

    def serialize_args(self) -> bytes:
        data = bytearray()
        data.extend(U8Type(self.value_dest_sreg).serialize())
        data.extend(U8Type(self.time_dest_sreg).serialize())
        data.extend(FwChanIdType(self.chan_id).serialize())
        return bytes(data)


@dataclass
class GetPrmDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.GET_PRM
    dest_sreg_index: int
    """U8: The local variable to store the parameter value in."""
    prm_id: int
    """FwPrmIdType: The parameter ID to get the value of."""

    def serialize_args(self) -> bytes:
        return (
            U8Type(self.dest_sreg_index).serialize()
            + FwPrmIdType(self.prm_id).serialize()
        )


@dataclass
class CmdDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.CMD
    op_code: int
    """FwOpcodeType: The opcode of the command."""
    arg_buf: bytes
    """[Fpy.MAX_SERIALIZABLE_REGISTER_SIZE] U8: The argument buffer of the command."""

    def serialize_args(self) -> bytes:
        data = bytearray()
        data.extend(FwOpcodeType(self.op_code).serialize())
        data.extend(self.arg_buf)
        return bytes(data)


@dataclass
class _DeserSerRegDirective(Directive):
    """
    Deserializes up to 8 bytes from a local variable into a register.
    """

    src_sreg_idx: int
    """U8: The local variable to deserialize from."""
    src_offset: int
    """FwSizeType: The starting offset to deserialize from."""
    dest_reg: int
    """U8: The destination register to deserialize into."""

    def serialize_args(self) -> bytes:
        data = bytearray()
        data.extend(U8Type(self.src_sreg_idx).serialize())
        data.extend(FwSizeType(self.src_offset).serialize())
        data.extend(U8Type(self.dest_reg).serialize())
        return bytes(data)


class DeserSerReg8Directive(_DeserSerRegDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.DESER_SER_REG_8


class DeserSerReg4Directive(_DeserSerRegDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.DESER_SER_REG_4


class DeserSerReg2Directive(_DeserSerRegDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.DESER_SER_REG_2


class DeserSerReg1Directive(_DeserSerRegDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.DESER_SER_REG_1


@dataclass
class SetRegDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.SET_REG

    dest: int
    """U8: The register to store the value in."""
    value: int
    """I64: The value to store in the register."""

    def serialize_args(self) -> bytes:
        return U8Type(self.dest).serialize() + I64Type(self.value).serialize()


@dataclass
class _BinaryRegOpDirective(Directive):
    lhs: int
    """U8: The left-hand side register for comparison."""
    rhs: int
    """U8: The right-hand side register for comparison."""
    res: int
    """U8: The destination register for the boolean result."""

    def serialize_args(self) -> bytes:
        data = bytearray()
        data.extend(U8Type(self.lhs).serialize())
        data.extend(U8Type(self.rhs).serialize())
        data.extend(U8Type(self.res).serialize())
        return bytes(data)


class OrDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.OR


class AndDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.AND


class IntEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.IEQ


class IntNotEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.INE


class UnsignedLessThanDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.ULT


class UnsignedLessThanOrEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.ULE


class UnsignedGreaterThanDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.UGT


class UnsignedGreaterThanOrEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.UGE


class SignedLessThanDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.SLT


class SignedLessThanOrEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.SLE


class SignedGreaterThanDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.SGT


class SignedGreaterThanOrEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.SGE


class FloatGreaterThanOrEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FGE


class FloatLessThanOrEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FLE


class FloatLessThanDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FLT


class FloatGreaterThanDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FGT


class FloatEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FEQ


class FloatNotEqualDirective(_BinaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FNE


@dataclass
class _UnaryRegOpDirective(Directive):
    src: int
    res: int

    def serialize_args(self) -> bytes:
        data = bytearray()
        data.extend(U8Type(self.src).serialize())
        data.extend(U8Type(self.res).serialize())
        return bytes(data)


@dataclass
class NotDirective(_UnaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.NOT


@dataclass
class FloatTruncateDirective(_UnaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FPTRUNC


@dataclass
class FloatExtendDirective(_UnaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FPEXT


@dataclass
class FloatToSignedIntDirective(_UnaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FPTOSI


@dataclass
class SignedIntToFloatDirective(_UnaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.SITOFP


@dataclass
class FloatToUnsignedIntDirective(_UnaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.FPTOUI


@dataclass
class UnsignedIntToFloatDirective(_UnaryRegOpDirective):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.UITOFP


@dataclass
class ExitDirective(Directive):
    opcode: ClassVar[DirectiveOpcode] = DirectiveOpcode.EXIT
    success: bool

    def serialize_args(self):
        return BoolType(self.success).serialize()


INT_EQUALITY_DIRECTIVES: dict[str, type[_BinaryRegOpDirective]] = {
    "==": IntEqualDirective,
    "!=": IntNotEqualDirective,
}

FLOAT_EQUALITY_DIRECTIVES: dict[str, type[_BinaryRegOpDirective]] = {
    "==": FloatEqualDirective,
    "!=": FloatNotEqualDirective,
}


INT_SIGNED_INEQUALITY_DIRECTIVES: dict[str, type[_BinaryRegOpDirective]] = {
    ">": SignedGreaterThanDirective,
    "<": SignedLessThanDirective,
    ">=": SignedGreaterThanOrEqualDirective,
    "<=": SignedLessThanOrEqualDirective,
}
INT_UNSIGNED_INEQUALITY_DIRECTIVES: dict[str, type[_BinaryRegOpDirective]] = {
    ">": UnsignedGreaterThanDirective,
    "<": UnsignedLessThanDirective,
    ">=": UnsignedGreaterThanOrEqualDirective,
    "<=": UnsignedLessThanOrEqualDirective,
}
FLOAT_INEQUALITY_DIRECTIVES: dict[str, type[_BinaryRegOpDirective]] = {
    ">": FloatGreaterThanDirective,
    "<": FloatLessThanDirective,
    ">=": FloatGreaterThanOrEqualDirective,
    "<=": FloatLessThanOrEqualDirective,
}

BINARY_COMPARISON_DIRECTIVES = {}
BINARY_COMPARISON_DIRECTIVES.update(INT_EQUALITY_DIRECTIVES)
BINARY_COMPARISON_DIRECTIVES.update(INT_SIGNED_INEQUALITY_DIRECTIVES)
BINARY_COMPARISON_DIRECTIVES.update(INT_UNSIGNED_INEQUALITY_DIRECTIVES)
BINARY_COMPARISON_DIRECTIVES.update(FLOAT_EQUALITY_DIRECTIVES)
BINARY_COMPARISON_DIRECTIVES.update(FLOAT_INEQUALITY_DIRECTIVES)



####
# default_settings.py:
#
# Contains default setup for the F prime flask application. Specifically, it is used to pass configuration
# down to the GDS config layers, and is used to specify a dictionary and packet spec for specifying
# the event, channels, and commands setup.
#
# Note: flask configuration is all done via Python files
#
####
import os

STANDARD_PIPELINE_ARGUMENTS = os.environ.get("STANDARD_PIPELINE_ARGUMENTS").split("|")

SERVE_LOGS = os.environ.get("SERVE_LOGS", "YES") == "YES"

MAX_CONTENT_LENGTH = 32 * 1024 * 1024  # Max length of request is 32MiB


# TODO: load real config



"""
@brief Command Data class

Instances of this class define a specific instance of a command with specific
argument values.

@data Created July 3, 2018
@author Josef Biberstein

@bug No known bugs
"""
import json

from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.bool_type import BoolType
from fprime.common.models.serialize.enum_type import EnumType
from fprime.common.models.serialize.numerical_types import (
    F32Type,
    F64Type,
    I8Type,
    I16Type,
    I32Type,
    I64Type,
    U8Type,
    U16Type,
    U32Type,
    U64Type,
)
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.string_type import StringType
from fprime.common.models.serialize.time_type import TimeBase, TimeType

from fprime_gds.common.data_types import sys_data


class CmdData(sys_data.SysData):
    """The CmdData class stores a specific command"""

    def __init__(self, cmd_args, cmd_temp, cmd_desc=None, cmd_time=None):
        """
        Constructor.

        Args:
            cmd_args: The arguments for the command. Should match the types of the
                      arguments in the cmd_temp object. Should be a tuple.
            cmd_temp: Command Template instance for this command (this provides
                      the opcode and argument types are stored)
            cmd_desc: command descriptor: Absolute/Relative. For sequences
            cmd_time: The time the command should occur. This is for sequences.
                      Should be a TimeType object with time base=TB_DONT_CARE

        Returns:
            An initialized CmdData object
        """
        super().__init__()
        self.id = cmd_temp.get_id()
        self.template = cmd_temp

        self.args, errors = self.process_args(cmd_args)
        self.time = cmd_time or TimeType(TimeBase["TB_DONT_CARE"].value)
        self.descriptor = cmd_desc

        # If any errors occur, then raise a aggregated error
        if [error for error in errors if error != ""]:
            raise CommandArgumentsException(errors)

    def get_template(self):
        """Get the template class associate with this specific data object

        Returns:
            Template -- The template class for this data object
        """

        return self.template

    def get_time(self):
        """ Return time """
        return self.time

    def get_descriptor(self):
        """ Return the descriptor """
        return self.descriptor

    def get_id(self):
        """Get the ID associate with the template of this data object

        Returns:
            An ID number
        """

        return self.id

    def get_arg_vals(self):
        """Get the values for each argument in a command.

        Returns:
            list -- a list of value objects that were used in this data object.
        """

        return [arg.val for arg in self.args]

    def get_args(self):
        """Get the arguments associate with the template of this data object

        Returns:
            list -- A list of type objects representing the arguments of the template of this data object (in order)
        """

        return self.args

    def get_str(self, time_zone=None, verbose=False, csv=False):
        """
        Convert the command data to a string

        Args:
            time_zone: (tzinfo, default=None) Timezone to print time in. If
                      time_zone=None, use local time.
            verbose: (boolean, default=False) Prints extra fields if True
            csv: (boolean, default=False) Prints each field with commas between
                                          if true

        Returns:
            String version of the command data
        """
        time_str = self.time.to_readable(time_zone)
        raw_time_str = str(self.time)
        name = self.template.get_full_name()

        if self.args is None:
            arg_str = "EMPTY COMMAND OBJ"
        else:
            # The arguments are currently serializable objects which cannot be
            # used to fill in a format string. Convert them to values that can be
            arg_val_list = self.get_arg_vals()
            arg_str = str(arg_val_list)

        if verbose and csv:
            return f"{time_str},{raw_time_str},{name},{self.id},{arg_str}"
        if verbose and not csv:
            return f"{time_str}: {name} ({self.id}) {raw_time_str} : {arg_str}"
        if not verbose and csv:
            return f"{time_str},{name},{arg_str}"
        return f"{time_str}: {name} : {arg_str}"

    def process_args(self, input_values):
        """ Process input arguments """
        errors = []
        args = []
        for val, arg_tuple in zip(input_values, self.template.arguments):
            try:
                arg_name, _, arg_type = arg_tuple
                arg_value = arg_type()
                self.convert_arg_value(val, arg_value)
                args.append(arg_value)
                errors.append("")
            except Exception as exc:
                errors.append(f"{arg_name}[{arg_type.__name__}]: {exc}")
        return args, errors

    @staticmethod
    def convert_arg_value(arg_val, arg_instance):
        if arg_val is None:
            raise CommandArgumentException(
                "Argument was not set"
            )
        if isinstance(arg_instance, BoolType):
            value = str(arg_val).lower().strip()
            if value in {"true", "yes"}:
                av = True
            elif value in {"false", "no"}:
                av = False
            else:
                raise CommandArgumentException("Argument value is not a valid boolean")
            arg_instance.val = av
        elif isinstance(arg_instance, EnumType):
            arg_instance.val = arg_val
        elif isinstance(arg_instance, (F64Type, F32Type)):
            arg_instance.val = float(arg_val)
        elif isinstance(
            arg_instance,
            (I64Type, U64Type, I32Type, U32Type, I16Type, U16Type, I8Type, U8Type),
        ):
            arg_instance.val = int(arg_val, 0) if isinstance(arg_val, str) else int(arg_val)
        elif isinstance(arg_instance, StringType):
            arg_instance.val = arg_val
        elif isinstance(arg_instance, (ArrayType, SerializableType)):
            arg_instance.val = json.loads(arg_val)
        else:
            raise CommandArgumentException(
                "Argument value could not be converted to type object"
            )

    def __str__(self):
        arg_str = "".join(f"{name} : {str(typ.val)} |" for name, typ in zip([arg[0] for arg in self.template.get_args()], self.args))
        arg_str = f"w/ args | {arg_str}"

        arg_info = f"{self.template.mnemonic} "

        return arg_info + arg_str if len(self.args) > 0 else arg_info


class CommandArgumentException(Exception):
    pass


class CommandArgumentsException(Exception):
    def __init__(self, errors):
        """
        Handle a list of errors as an exception.
        """
        super().__init__(" ".join(errors))
        self.errors = errors



""" errors.py:

Contains the necessary definitions to handle errors across flask in a standard way. This will hopefully make the GDS
more stable and less likely to fail on errors that occur.

@author lestarch
"""

from flask import Flask, jsonify
from flask_restful import Api


def build_error_object(error, args=None):
    """Builds an error object from an error"""
    return {
        "type": str(type(error).__name__),
        "message": str(error),
        "args": args if args is not None else [],
    }


def handle_flask_error(error):
    """Handle an error within the flask usage

    Handling an error within flask should produce a standard error message and pass it back to the JavaScript
    application. The error response is expected to look like:

        {"errors": [{"type": "<error type>, "message": "<display message>"}, ...]}

    This will include setting the status code to 500. It is up to any specific resource to continue within the context
    of a recoverable error and the function build_error_object may be called to build an error message.

    Args:
        error: the error to be reported

    Returns:
        response JSON, status code
    """
    status_code = getattr(error, "code", 500)
    response = {"errors": [build_error_object(error, getattr(error, "args", []))]}
    return jsonify(response), status_code


class ErrorHandlingApi(Api):
    """Subclass of the flask_restful API only to change the error handle"""

    def handle_error(self, error):
        """Handle errors within flask_restful by overriding the handle_error method"""
        return handle_flask_error(error)


def setup_error_handling(app: Flask):
    """Setup the error handling for flask and get a flask_restful API with similar error handling

    Sets up a flask_restful API that will handle errors in the standard way and registers the same error handling to
    the flask app for non-restful errorrs

    Args:
        app: flask application to register errors

    Returns:
        flask restful api with standardized error handling
    """
    app.errorhandler(Exception)(handle_flask_error)
    return Api(app)



####
# commands.py:
#
# This file provides the API for the commands Gds interface for use with the Gds Flask server. This
# API should provide the following HTML API behaviors:
#
#  GET /commands: list all commandsi history available to the GUI. Note: this also provides a full
#                 command listing.
#  PUT /commands/<command>: issue a command through the GDS
#      Data: {
#                "key": "0xfeedcafe", # A key preventing accidental issuing of a command
#                "args": {
#                            <arg-key>:<arg-value>,
#                             ...
#                             ...
#                        }
#             }
#
#
# Note: for commands, these are not true "REST" objects, and thus this is a bit of a stretch to use
#       a restful interface here. It is done this way to be in-tandem with the events and telemetry
#       APIs for maintainability.
####
import flask_restful
import flask_restful.reqparse
import fprime.common.models.serialize.type_exceptions
import werkzeug.exceptions

import fprime_gds.common.data_types.cmd_data
from fprime_gds.flask.resource import DictionaryResource, HistoryResourceBase


class CommandDictionary(DictionaryResource):
    """Channel dictionary shares implementation"""


class CommandHistory(HistoryResourceBase):
    """Command history requires no deviation from the base history implementation"""


class MissingArgumentException(werkzeug.exceptions.BadRequest):
    """Did not supply an argument"""

    def __init__(self):
        super().__init__("Did not supply all required arguments.")


class CommandArgumentsInvalidException(werkzeug.exceptions.BadRequest):
    """Command arguments failed to validate properly"""

    def __init__(self, errors):
        super().__init__(f"Failed to validate all arguments: {', '.join(errors)}")
        self.args = errors


class InvalidCommandException(werkzeug.exceptions.BadRequest):
    """Requested invalid command"""

    def __init__(self, key):
        super().__init__(f"{ key } is not a valid command")


class Command(flask_restful.Resource):
    """
    Command object used to send commands into the GDS.
    """

    def __init__(self, sender):
        """
        Constructor: setup the parser for incoming command runs
        """
        self.parser = flask_restful.reqparse.RequestParser()
        self.parser.add_argument(
            "key", required=True, help="Protection key. Must be: 0xfeedcafe."
        )
        self.parser.add_argument(
            "arguments", action="append", help="Argument list to pass to command."
        )
        self.sender = sender

    def put(self, command):
        """
        Receive a command run request.
        @param command: command identification
        """
        args = self.parser.parse_args()
        key = args.get("key", None)
        arg_list = args.get("arguments", [])
        # Error checking
        if key is None or int(key, 0) != 0xFEEDCAFE:
            flask_restful.abort(
                403,
                message=f"{key} is invalid command key. Supply 0xfeedcafe to run command.",
            )
        if arg_list is None:
            arg_list = []
        try:
            self.sender.send_command(command, arg_list)
        except fprime.common.models.serialize.type_exceptions.NotInitializedException:
            raise MissingArgumentException()
        except fprime_gds.common.data_types.cmd_data.CommandArgumentsException as exc:
            raise CommandArgumentsInvalidException(exc.errors)
        except KeyError as key_error:
            raise InvalidCommandException(key_error)
        return {"message": "success"}



"""
test_logger.py:

A wrapper on the openpyxl library that provides the ability to log messages in an excel sheet. Used
by the test API to log events, asserts, test cases and user messages. The documentation for openpyxl
can be found here:
https://openpyxl.readthedocs.io/en/stable/index.html

If the openpyxl library isn't installed, this class does nothing.

This class uses a write-only optimization that should allow for creating large log files without
hogging too much memory. Write-only optimization can be found here:
https://openpyxl.readthedocs.io/en/stable/optimized.html#write-only-mode

:author: koran
"""
import datetime
import os
import threading
import time

# If openpyxl isn't installed, ignore all functionality in this module
try:
    from openpyxl import Workbook
    from openpyxl.cell import WriteOnlyCell
    from openpyxl.styles import Alignment, Font, PatternFill
    from openpyxl.utils.exceptions import WorkbookAlreadySaved

    MODULE_INSTALLED = True
except ImportError:
    MODULE_INSTALLED = False


class TestLogger:
    """
    User-accessible colors. Can be used for the color arguments
    """

    __test__ = False
    BROWN = "E6CCB3"
    RED = "FF9999"
    ORANGE = "FFCC99"
    YELLOW = "FFFF99"
    GREEN = "ADEBAD"
    BLUE = "99CCFF"
    PURPLE = "CC99FF"
    GRAY = "D9D9D9"
    WHITE = "FFFFFF"

    """
    User-accessible styles. Must be used for the style arguments
    """
    BOLD = "BOLD"
    ITALICS = "ITALICS"
    UNDERLINED = "UNDERLINED"

    if MODULE_INSTALLED:
        __align = Alignment(vertical="top", wrap_text=True)
        __font_name = "calibri"
        __time_fmt = "%H:%M:%S.%f"

    def __init__(self, output_path, time_format=None, font_name=None):
        """
        Constructs a TestLogger

        Args:
            output_path: a path where log files will
            time_format: an optional string to specify the timestamp format. See datetime.strftime
            font_name: an optional string to specify the font
        """
        if not MODULE_INSTALLED:
            return

        if not isinstance(output_path, str):
            raise TypeError(
                "Test Logger requires a filename where the output can be saved."
            )
        self.start_time = time.time()
        date_string = datetime.datetime.fromtimestamp(self.start_time).strftime(
            "%Y-%m-%dT%H:%M:%S"
        )
        self.filename = os.path.join(output_path, f"TestLog_{date_string}.xlsx")
        self.workbook = Workbook(write_only=True)
        self.worksheet = self.workbook.create_sheet()
        self.ws_saved = False

        self.time_format = self.__time_fmt if time_format is None else time_format
        self.font_name = self.__font_name if font_name is None else font_name

        timestring = datetime.datetime.fromtimestamp(self.start_time).strftime(
            self.time_format
        )
        self.worksheet.column_dimensions["A"].width = len(timestring) + 1
        self.worksheet.column_dimensions["D"].width = 120

        date_string = datetime.datetime.fromtimestamp(self.start_time).strftime(
            "%H:%M:%S.%f on %m/%d/%Y"
        )
        top = [self.__get_cell(f"Test began at {date_string}")]
        self.worksheet.append(top)

        labels = ["Log Time", "Case ID", "Sender", "Message"]
        header = self.__get_ws_row(labels, style=self.BOLD)
        self.worksheet.append(header)

        self.case_id = "NA"

        self.lock = threading.Lock()

    def log_message(self, message, sender="NA", color=None, style=None, case_id=None):
        """
        Logs a message to the TestLog. Each message will include a timestamp beforehand.
        Note: Once specified, the test case's case_id will persist in the logs until it is
        specified again.

        Args:
            message: a message to log (str).
            sender: a short string describing who created the log message
            color: a string object containing a color hex code "######"
            style: a string choosing 1 of 3 formatting options (ITALICS, BOLD, UNDERLINED)
            case_id: a short identifier to denote which test case the log message belongs to
        """
        if not MODULE_INSTALLED:
            return

        ts = time.time()
        timestring = datetime.datetime.fromtimestamp(ts).strftime(self.time_format)

        if case_id is not None:
            if not isinstance(case_id, str):
                case_id = str(case_id)
            self.case_id = case_id

        strings = [timestring, self.case_id, sender, message]
        self.lock.acquire()
        try:
            print(f"{timestring} [{sender}] {message}")
            if not self.ws_saved:
                row = self.__get_ws_row(strings, color, style)
                self.worksheet.append(row)
        except WorkbookAlreadySaved:
            self.ws_saved = True
            print(f"{timestring} [TestLogger] Workbook has already been saved.")
        finally:
            self.lock.release()

    def close_log(self):
        """
        Saves the write-only workbook. Should be called only once when the log is completed.
        """
        if not MODULE_INSTALLED:
            return

        self.workbook.save(filename=self.filename)
        self.ws_saved = True

    def __get_cell(self, string, color=None, style=None):
        """
        Helper method for log message. This method takes a string as well as color and style
        arguments to create a write-only cell.

        Args:
            string: a string object to be written to the cell's value field.
            color: a string object containing a color hex code "######"
            style: a string choosing 1 of 3 formatting options (ITALICS, BOLD, UNDERLINED)
        """
        if not MODULE_INSTALLED:
            return

        cell = WriteOnlyCell(self.worksheet, value=string)
        if color is not None:
            # pylint: disable=E0237
            cell.fill = PatternFill("solid", fgColor=color)
        # pylint: disable=E0237
        cell.font = Font(
            name=self.font_name,
            bold=(style == self.BOLD),
            italic=(style == self.ITALICS),
            underline=("single" if style == self.UNDERLINED else "none"),
        )
        # pylint: disable=E0237
        cell.alignment = self.__align
        return cell

    def __get_ws_row(self, strings, color=None, style=None):
        return [self.__get_cell(string, color, style) for string in strings]



"""
@brief Command Template class

Instances of this class describe a specific command type.

@data Created July 2, 2018
@author Josef Biberstein

@bug No known bugs
"""

import copy

from fprime.common.models.serialize.type_base import BaseType
from fprime.common.models.serialize.type_exceptions import (
    ArgLengthMismatchException,
    TypeMismatchException,
)

from . import data_template


class CmdTemplate(data_template.DataTemplate):
    def __init__(self, opcode, mnemonic, component, arguments, description=None):
        """
        Constructor

        Args:
            opcode: The cmd's opcode (integer)
            mnemonic: The mnemonic used to refer to the cmd (string)
            component: The name of the component that produces the cmd (string)
            arguments: List of arguments in tuple form. Each tuple should be:
                       (arg name, arg description, arg obj). Where arg obj is an
                       object of a type derived from the class Base Type.
                       Arg description may be None.
            description: (Optional) Description of the cmd (string)
        """
        super().__init__()
        # Make sure correct types are passed
        if not isinstance(component, str):
            raise TypeMismatchException(str, type(component))

        if not isinstance(mnemonic, str):
            raise TypeMismatchException(str, type(mnemonic))

        if not isinstance(opcode, int):
            raise TypeMismatchException(int, type(opcode))

        if description is not None and not isinstance(description, str):
            raise TypeMismatchException(str, type(description))

        if not isinstance(arguments, list):
            raise TypeMismatchException(list, type(arguments))

        for (argname, argdesc, argtype) in arguments:
            #
            if not isinstance(argname, str):
                raise TypeMismatchException(int, type(argname))
            #
            if argdesc is not None and not isinstance(argdesc, str):
                raise TypeMismatchException(int, type(argdesc))
            #
            if not issubclass(argtype, BaseType):
                raise TypeMismatchException(BaseType, argtype)

        # Initialize command internal variables
        self.comp_name = component
        self.mnemonic = mnemonic
        self.opcode = opcode
        self.description = description
        self.arguments = arguments
        self.descriptor = None
        self.sec = None
        self.usec = None

    def get_full_name(self):
        """
        Get the full name of this event

        Returns:
            The full name (component.channel) for this event
        """
        return f"{self.comp_name}.{self.mnemonic}"

    def get_comp_name(self):
        return self.comp_name

    def get_mnemonic(self):
        return self.mnemonic

    def get_op_code(self):
        return self.opcode

    def get_id(self):
        """
        Get the Command's id, which is its opcode.

        This function is implemented to maintain the sys_data interface

        Returns:
            The command's opcode
        """
        return self.opcode

    def get_name(self):
        """
        Get the Command's name, which is its mnemonic

        This function is implemented to maintain the sys_data interface

        Returns:
            The command's mnemonic
        """
        return self.mnemonic

    def get_description(self):
        return self.description

    def get_args(self):
        """
        Returns a list of argument information

        Returns:
            A list of tuples where each tuple represents an argument. Each tuple
            in the form: (arg name, arg description, arg obj). Where arg obj is
            an object of a type derived from the class Base Type. Arg
            description may be None.
        """
        return self.arguments

    # Methods to keep compatibility with the names used in
    # Gds/src/fprime_gds/common/encoders/seq_writer. This allows sharing
    # of the code.

    def setArgs(self, values):

        if len(values) != len(self.arguments):
            raise ArgLengthMismatchException(len(self.arguments), len(values))

        # set the values of the arguments
        new_arg_list = []
        for value, (arg_name, arg_desc, arg_value) in zip(values, self.arguments):
            new_value = copy.deepcopy(arg_value)
            new_value.val = value
            new_arg_list.append((arg_name, arg_desc, new_value))

        self.arguments = new_arg_list

    def setDescriptor(self, desc):
        self.descriptor = desc

    def setSeconds(self, sec):
        self.sec = sec

    def setUseconds(self, usec):
        self.usec = usec

    def getOpCode(self):
        return self.opcode

    def getDescriptor(self):
        """
        Returns the time descriptor, relative or absolute

        Returns:
            A integer that indicates whether the command is relative or absolute
        """
        return self.descriptor

    def getSeconds(self):
        return self.sec

    def getUseconds(self):
        return self.usec

    def getArgs(self):
        return self.get_args()

    def __repr__(self):
        arg_strs = []
        for arg in self.arguments:
            arg_strs.append(arg[0] + ": " + str(arg[2]))
            
        args_str = ", ".join(arg_strs)
        return f"CmdTemplate({self.comp_name}.{self.mnemonic}, args: ({args_str}))"



"""
base.py:

This file specifies the base-adapter for the F prime comm-layer. This class defines the basic methods needed to interact
with various wire formats as they are supported by the F prime comm-layer. This file defines a single abstract base
class representing the core features of the adapter class that must be implemented by every implementation of the
adapter for use with the comm-layer.

@author lestarch
"""
import abc
from typing import Type
from fprime_gds.plugin.definitions import gds_plugin_implementation, gds_plugin_specification


class BaseAdapter(abc.ABC):
    """
    Base adapter for adapting the communications layer. This essentially breaks down to providing the ability to read
    data from, and write to the necessary wire-format. The children of this class must at least implement the 'read' and
    'write' functions to ensure that data can be read and written. 'open' and 'close' are also provided as a helper to
    the subclass implementer to place resource initialization and release code, however; these implementations are
    defaulted not overridden.
    """

    def open(self):
        """Null default implementation"""

    def close(self):
        """Null default implementation"""

    @abc.abstractmethod
    def read(self, timeout=0.500):
        """
        Read from the interface. Must be overridden by the child adapter. Throw no fatal errors, reconnect instead. This
        call is expected to block waiting on incoming data.

        :param size: maximum size of data to read before breaking
        :param timeout: timeout for the block, default: 0.500 (500ms) as blocking w/o timeout may be uninterruptible
        :return: byte array of data, or b'' if no data was read
        """

    @abc.abstractmethod
    def write(self, frame):
        """
        Write to the interface. Must be overridden by the child adapter. Throw no fatal errors, reconnect instead.

        :param frame: framed data to uplink
        :return: True if data sent through adapter, False otherwise
        """

    @classmethod
    @gds_plugin_specification
    def register_communication_plugin(cls) -> Type["BaseAdapter"]:
        """Register a communications adapter

        Plugin hook for registering a plugin that supplies an adapter to the communications interface (radio, uart, i2c,
        etc). This interface is expected to read and write bytes from a wire and will be provided to the framing system.

        Note: users should return the class, not an instance of the class. Needed arguments for instantiation are
        determined from class methods, solicited via the command line, and provided at construction time to the chosen
        instantiation.

        Returns:
            BaseAdapter subclass
        """
        raise NotImplementedError()


class NoneAdapter(BaseAdapter):
    """ None adapter used to turn off the comm script """

    @classmethod
    def get_name(cls):
        """ Get name of the non-adapter """
        return "none"

    def read(self, timeout=0.500):
        """ Raise exception if this is called"""
        raise NotImplementedError()

    def write(self, frame):
        """ Raise exception if this is called"""
        raise NotImplementedError()

    @classmethod
    @gds_plugin_implementation
    def register_communication_plugin(cls):
        """ Register this as a plugin """
        return cls


import filecmp
from pathlib import Path
import tempfile
import unittest

import fprime_gds.common.tools.seqgen as seqgen

class APITestCases(unittest.TestCase):
    def test_nominal_sequence(self):
        self.assertEqual(self.check_sequence_generates_expected_binary(
            Path(__file__).parent / "input" / "simple_sequence.seq",
            Path(__file__).parent / "expected" / "simple_expected.bin",
            Path(__file__).parent / "resources" / "simple_dictionary.json"
        ), True)
    
    def test_fail_unmatched_command_sequence(self):
        with self.assertRaisesRegex(seqgen.SeqGenException, "does not match any command in the command dictionary."):
            self.check_sequence_generates_expected_binary(
                Path(__file__).parent / "input" / "simple_bad_sequence.seq",
                Path(__file__).parent / "expected" / "simple_expected.bin",
                Path(__file__).parent / "resources" / "simple_dictionary.json"
            )
        
    def check_sequence_generates_expected_binary(self, 
                                                 input_sequence, 
                                                 expected_binary, 
                                                 dictionary):
        temp_dir = tempfile.TemporaryDirectory()
        output_bin = Path(f"{temp_dir.name}/out_binary")
        seqgen.generateSequence(
            input_sequence,
            output_bin,
            dictionary,
            0xffff
        )
        is_equal = filecmp.cmp(output_bin, expected_binary)
        temp_dir.cleanup()
        return is_equal
        
    
if __name__ == "__main__":
    unittest.main()


"""
@brief Class to store data from each specific file_data type

@date Created July 10, 2019
@author Blake A. Harriman

@bug No known bugs
"""
from enum import Enum

from fprime_gds.common.data_types import sys_data


class FilePacketType(Enum):
    """
    An enumeration of available file packets
    """

    START = 0
    DATA = 1
    END = 2
    CANCEL = 3


# Takes care of the START packets that the file_decoder will receive
class StartPacketData(sys_data.SysData):
    """
    The StartPacketData class stores the start packet information
    """

    # Initializes the start packet data with all of the necessary information
    def __init__(self, seqID, size, sourcePath, destPath):
        """
        Constructor.

        Args:
            packetType: The type of packet the object is
            seqID: The ID that identifies this specific object being created
            size: The size of the object
            lengthSP: The length of the source path
            sourcePath: The path that the file is coming from
            lengthDP: The length of the destination path
            destPath: The path that the file is ending up at

        Returns:
            An initialized StartPacketData object
        """
        super().__init__()
        self.packetType = FilePacketType.START
        self.seqID = seqID
        self.size = size
        # self.lengthSP = lengthSP
        self.sourcePath = sourcePath
        # self.lengthDP = lengthDP
        self.destPath = destPath


# Takes care of the DATA packets that the file_decoder will receive
class DataPacketData(sys_data.SysData):
    """
    The DataPacketData class stores the data packet information
    """

    # Initializes the data packet data with all of the necessary information
    def __init__(self, seqID, offset, dataVar):
        """
        Constructor.

        Args:
            packetType: The type of packet the object is
            seqID: The ID that identifies this specific object being created
            offset: How far off the object is from the original location
            length: The length of the object
            dataVar: Contains the rest of the data for the object

        Returns:
            An initialized DataPacketData object
        """
        super().__init__()
        self.packetType = FilePacketType.DATA
        self.seqID = seqID
        self.offset = offset
        self.dataVar = dataVar
        self.length = len(self.dataVar)


# Takes care of the END packets that the file_decoder will receive
class EndPacketData(sys_data.SysData):
    """
    The EndPacketData class stores the end packet information
    """

    # Initializes the end packet data with all of the necessary information
    def __init__(self, seqID, hashValue):
        """
        Constructor.

        Args:
            packetType: The type of packet the object is
            seqID: The ID that identifies this specific object being created
            hashValue: THe 32 bit has value described in the CFPD protocol

        Returns:
            An initialized EndPacketData object
        """
        super().__init__()
        self.packetType = FilePacketType.END
        self.seqID = seqID
        self.hashValue = hashValue


# Takes care of the CANCEL packets that the file_decoder will receive
class CancelPacketData(sys_data.SysData):
    """
    The CancelPacketData class stores the cancel packet information
    """

    # Initializes the cancel packet data with all of the necessary information
    def __init__(self, seqID):
        """
        Constructor.

        Args:
            packetType: The type of packet the object is
            seqID: The ID that identifies this specific object being created

        Returns:
            An initialized StartPacketData object
        """
        super().__init__()
        self.packetType = FilePacketType.CANCEL
        self.seqID = seqID



"""
Handles executing the "channels" CLI command for the GDS
"""

from typing import Iterable

import fprime_gds.common.gds_cli.test_api_utils as test_api_utils
from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.gds_cli.base_commands import QueryHistoryCommand
from fprime_gds.common.pipeline.dictionaries import Dictionaries
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.testing_fw.api import IntegrationTestAPI


class ChannelsCommand(QueryHistoryCommand):
    """
    Takes in the given arguments and prints an appropriate formatted string of
    recent event data that matches the user's criteria.

    For descriptions of these arguments, and more function details, see the
    events command definition at:
        Gds/src/fprime_gds/executables/fprime_cli.py
    """

    @classmethod
    def _get_item_list(
        cls,
        project_dictionary: Dictionaries,
        filter_predicate: predicates.predicate,
    ) -> Iterable[ChData]:
        """
        Gets a list of open telemetry channels in the system and prints their
        details out in an ID-sorted list.

        :param project_dictionary: The dictionary object for the project
            containing the channel type definitions
        :param filter_predicate: Test API predicate used to filter shown
            channels
        :return: List of channels that passed the filter
        """
        channel_list = test_api_utils.get_item_list(
            item_dictionary=project_dictionary.channel_id,
            search_filter=filter_predicate,
            template_to_data=ChData.get_empty_obj,
        )
        return channel_list

    @classmethod
    def _get_upcoming_item(
        cls,
        api: IntegrationTestAPI,
        filter_predicate: predicates.predicate,
        min_start_time="NOW",
        timeout: float = 5,
    ) -> ChData:
        """
        Returns the next upcoming channel data after the given time that
        matches the given filter.
        """
        return test_api_utils.get_upcoming_channel(
            api, filter_predicate, min_start_time, timeout
        )



"""
A collection of base classes used for the backend implementations of the GDS
CLI commands
"""

import abc
import json
import sys
from typing import Iterable

import fprime_gds.common.gds_cli.filtering_utils as filtering_utils
import fprime_gds.common.gds_cli.test_api_utils as test_api_utils
from fprime_gds.common.pipeline.dictionaries import Dictionaries
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.testing_fw.api import IntegrationTestAPI
from fprime_gds.executables.cli import StandardPipelineParser


class BaseCommand(abc.ABC):
    """
    The base class for implementing a GDS CLI channel/event/command-send functionality
    Subclasses must implement the _get_item_list and handle_command methods
    """

    ####################################################################
    #   Abstract Methods
    ####################################################################

    @classmethod
    @abc.abstractmethod
    def _execute_command(cls, args, api: IntegrationTestAPI):
        """
        Command logic of sending/receiving channels/commands/events
        """

    @classmethod
    @abc.abstractmethod
    def _get_item_list(
        cls,
        project_dictionary: Dictionaries,
        filter_predicate: predicates.predicate,
    ) -> Iterable:
        """
        Gets a sorted list of items from the dictionary, filtered by the given
        predicate, and return it.

        :param project_dictionary: The dictionary object for the project
            containing the item type definitions
        :param filter_predicate: Test API predicate used to filter shown items
        :return: An iterable collection of items that passed the filter
        """

    @classmethod
    @abc.abstractmethod
    def _get_item_string(
        cls,
        item,
        json: bool = False,
    ) -> str:
        """
        Takes an F' item and returns a human-readable string of its data.

        :param item: The F' item to convert to a string
        :param json: Whether to print out each item in JSON format or not
        :return: A string representation of "item"
        """

    ####################################################################
    #   Shared Methods
    ####################################################################
    @classmethod
    def _get_item_list_string(
        cls,
        items: Iterable,
        json: bool = False,
    ) -> str:
        """
        Takes a list of items from the dictionary and returns a human-readable
        string containing their details.

        :param items: An iterable collection of F' data objects
        :param json: Whether to print out each item in JSON format or not
        :return: A string containing the details of each item
        """
        return "\n".join(cls._get_item_string(item, json) for item in items)

    @classmethod
    def _get_search_filter(
        cls,
        ids: Iterable[int],
        components: Iterable[str],
        search: str,
        json: bool = False,
    ) -> predicates.predicate:
        """
        Returns a predicate that can be used to filter any received messages.
        This is done to link printing code to filtering, so that filtering ALWAYS
        filters the same strings as are printed.

        NOTE: Currently assumes that item list strings will contain the
        individual strings of each items to work with both, which there's no
        guarantee of if _get_item_list_string is overridden

        :param search: The F' item to convert to a string
        :param json: Whether to convert each item to a JSON representation
            before filtering them
        :return: A string representation of "item"
        """

        def item_to_string(x):
            return cls._get_item_string(x, json)

        return filtering_utils.get_full_filter_predicate(
            ids, components, search, to_str=item_to_string
        )

    @classmethod
    def _list_all_possible_items(
        cls, dictionary_path: str, search_filter: predicates.predicate, json: bool
    ) -> str:
        """
        Returns a string of the information for all relevant possible items
        for this command, filtered using the given predicate.

        :param dictionary_path: The string path to the dictionary file we should
            look for possible items in
        :param search_filter: The predicate to filter the items with
        :param json: Whether to print out each item in JSON format or not
        :return: A string describing each item relevant to this command that
            passes the given filter
        """
        project_dictionary = Dictionaries()
        project_dictionary.load_dictionaries(dictionary_path, packet_spec=None, packet_set_name=None)
        items = cls._get_item_list(project_dictionary, search_filter)
        return cls._get_item_list_string(items, json)

    @classmethod
    def handle_arguments(cls, args, **kwargs):
        """
        Entrypoint for the command. Handles parsing the arguments and creating the
        StandardPipeline and Integration API, then calls the execution of the command logic.
        Finally, tears down the pipeline and API.
        """
        pipeline = None
        api = None
        try:
            # Parsing the arguments
            pipeline_parser = StandardPipelineParser()
            pipeline_parser.handle_arguments(args, **kwargs, client=True)

            # If the user is just listing all possible items, do that and exit
            if args.is_printing_list:
                search_filter = cls._get_search_filter(
                    args.ids, args.components, args.search, args.json
                )
                cls._log(
                    cls._list_all_possible_items(
                        args.dictionary, search_filter, args.json
                    )
                )
                return

            # Set up StandardPipeline and Integration API
            pipeline = pipeline_parser.pipeline_factory(args)
            api = IntegrationTestAPI(pipeline)
            api.setup()

            # Execute the command logic
            cls._execute_command(args, api)

        # Tear down resources
        finally:
            try:
                if api is not None:
                    api.teardown()
            except Exception as exc:
                print(f"[WARNING] Exception in API teardown: {exc}", file=sys.stderr)
            try:
                if pipeline is not None:
                    pipeline.disconnect()
            except Exception as exc:
                print(
                    f"[WARNING] Exception in pipeline teardown: {exc}", file=sys.stderr
                )

    @classmethod
    def _log(cls, log_text: str):
        """
        Takes the given string and logs it (by default, logs all output to the
        console). Will ignore empty strings.

        :param log_text: The string to print out
        """
        if log_text:
            print(log_text)
            sys.stdout.flush()


class QueryHistoryCommand(BaseCommand):
    """
    The base class for a set of related GDS CLI commands that need to query and
    display received data from telemetry channels and F' events.
    """

    @classmethod
    @abc.abstractmethod
    def _get_upcoming_item(
        cls,
        api: IntegrationTestAPI,
        filter_predicate: predicates.predicate,
        min_start_time="NOW",
        timeout: float = 5.0,
    ):
        """
        Retrieves an F' item that has occurred since the given time and returns
        its data.
        """
    

    @classmethod
    def _get_item_string(
        cls,
        item: "SysData",
        as_json: bool = False,
    ) -> str:
        """
        Takes an F' item and returns a human-readable string of its data.

        :param item: EventData or ChData
        :param as_json: Whether to print out each item in JSON format or not
        :return: A string representation of "item"
        """
        if not item:
            return ""
        return json.dumps(item.get_dict()) if as_json else item.get_str(verbose=True)

    @classmethod
    def _execute_command(cls, args, api: IntegrationTestAPI):
        """
        Logic for receiving events/channels given the parsed arguments.
        This differentiates between the events and channels because their
        implementation of _get_upcoming_item is different.
        """
        search_filter = cls._get_search_filter(
            args.ids, args.components, args.search, args.json
        )
        # Timeout <= 0 means we should keep going until interrupted
        if args.timeout > 0:
            item = cls._get_upcoming_item(api, search_filter, "NOW", args.timeout)
            cls._log(cls._get_item_string(item, args.json))
        else:

            def print_upcoming_item(min_start_time="NOW"):
                item_ = cls._get_upcoming_item(api, search_filter, min_start_time)
                cls._log(cls._get_item_string(item_, args.json))
                # Update time so we catch the next item since the last one
                if item_:
                    min_start_time = predicates.greater_than(item_.get_time())
                    min_start_time = filtering_utils.time_to_data_predicate(
                        min_start_time
                    )
                return (min_start_time,)

            test_api_utils.repeat_until_interrupt(print_upcoming_item, "NOW")



"""
files.py:

Module to encapsulate the handling of file uplink and downlink. This includes setting up and registering both the file
uplinker and the file downlinker. These will handle the transferring of file data to/from the host into the
communications layer.

@author mstarch
"""
from pathlib import Path
import fprime_gds.common.files.downlinker
import fprime_gds.common.files.uplinker


class Filing:
    """
    A composition object that holds references to a FileUplinker and file downlinker.
    """

    def __init__(self):
        """
        The filing composition consists of an uplinker and a downlinker
        """
        self.__uplinker = None
        self.__downlinker = None

    def setup_file_handling(
        self, down_store, file_encoder, file_decoder, distributor, log_dir
    ):
        """
        Sets up the file handling (uplink and downlink) from a pair of encoders and decoders.
        Raises a PermissionError if the down_store is not writable.

        :param down_store: downlink storage directory
        :param file_encoder: file encoder for uplink
        :param file_decoder: file decoder for downlink
        :param distributor: data distributor to register handshaking to
        :param log_dir: log directory to output downlink logs
        """
        self.__uplinker = fprime_gds.common.files.uplinker.FileUplinker(file_encoder)
        self.__downlinker = fprime_gds.common.files.downlinker.FileDownlinker(
            down_store, log_dir=log_dir
        )
        file_decoder.register(self.__downlinker)
        distributor.register("FW_PACKET_HAND", self.__uplinker)
        try:
            Path(down_store).mkdir(parents=True, exist_ok=True)
        except PermissionError:
            raise PermissionError(
                f"{down_store} is not writable. Fix permissions or change storage directory with --file-storage-directory."
            )

    @property
    def uplinker(self):
        """
        Property to return the uplinker

        :return: uplinker object
        """
        return self.__uplinker

    @property
    def downlinker(self):
        """
        Property to return the downlinker

        :return: downlinker object
        """
        return self.__downlinker



"""fprime_gds.executables.apps: an implementation of start-up apps in fprime

There are twp ways to approach start=up applications in fprime. First, is to implement a run method via a subclass of
`GdsFunction`. This gives the implementor the ability to run anything within the run function that python offers,
however; this comes with complexity of setting up a new thread/process/isolation to ensure that the plugin does not
threaten the fprime-gds core functionality and processes.

The second method is to inherit from `GdsApp` implementing the `get_process_invocation` function to return the necessary
command line that will be spun into its own process.

@author lestarch
"""

import subprocess
import sys
from abc import ABC, abstractmethod
from argparse import Namespace
from typing import final, List, Dict, Tuple, Type

from fprime_gds.plugin.definitions import gds_plugin_specification, gds_plugin
from fprime_gds.plugin.system import Plugins
from fprime_gds.executables.cli import (
    CompositeParser,
    ParserBase,
    BareArgumentParser,
    StandardPipelineParser,
    PluginArgumentParser,
)
from fprime_gds.common.pipeline.standard import StandardPipeline


class GdsBaseFunction(ABC):
    """Base functionality for pluggable GDS start-up functions

    GDS start-up functionality is pluggable. This class acts as a base for pluggable functionality supplies helpers to
    the various start-up plugins.

    Developers who intend to run in an isolated subprocess are strongly encouraged to use `GdsApp` (see below).
    Developers who need flexibility may use GdsFunction.
    """

    @abstractmethod
    def run(self):
        """Run the start-up function

        Run the start-up function unconstrained by the limitations of running in a dedicated subprocess.

        """
        raise NotImplementedError()


class GdsFunction(GdsBaseFunction, ABC):
    """Functionality for pluggable GDS start-up functions

    GDS start-up functionality is pluggable. This class acts as a wide-open implementation of functionality via a single
    `run` callback. Developers have complete control of the start-up functionality. However, this comes at the cost of
    instability in that case of poorly designed functions.

    Developers who intend to run in an isolated subprocess are strongly encouraged to use `GdsApp` (see below).

    Plugin developers are required to implement a single function `run`, which must take care of setting up and running
    the start-up function. Developers **must** handle the isolation of this functionality including spinning off a new
    thread, subprocess, etc. Additionally, the developer must define the `register_gds_function_plugin` class method
    annotated with the @gds_plugin_implementation annotation.

    Standard plug-in functions (get_name, get_arguments) are available should the implementer desire these features.
    Arguments will be supplied to the class's `__init__` function.
    """

    @classmethod
    @gds_plugin_specification
    def register_gds_function_plugin(cls) -> Type["GdsFunction"]:
        """Register gds start-up functionality

        Plugin hook for registering a plugin that supplies start-up functionality. This functionality will run on start-up
        of the GDS network.

        Note: users should return the class, not an instance of the class. Needed arguments for instantiation are
        determined from class methods, solicited via the command line, and provided at construction time to the chosen
        instantiation.

        Returns:
            GDSFunction subclass
        """
        raise NotImplementedError()


class GdsApp(GdsBaseFunction):
    """GDS start-up process functionality

    A pluggable base class used to start a new process as part of the GDS command line invocation. This allows
    developers to add process-isolated functionality to the GDS network.

    Plugin developers are required to implement the `get_process_invocation` function that returns a list of arguments
    needed to invoke the process via python's `subprocess`. Additionally, the developer must define the
    `register_gds_function_plugin` class method annotated with the @gds_plugin_implementation annotation.

    Standard plug-in functions (get_name, get_arguments) are available should the implementer desire these features.
    Arguments will be supplied to the class's `__init__` function.
    """

    def __init__(self, **arguments):
        """Construct the communication applications around the arguments

        Command line arguments are passed in to match those returned from the `get_arguments` functions.

        Args:
            arguments: arguments from the command line
        """
        self.process = None
        self.arguments = arguments

    def run(self):
        """Run the application as an isolated process

        GdsFunction objects require an implementation of the `run` command. This implementation will take the arguments
        provided from `get_process_invocation` function and supplies them as an invocation of the isolated subprocess.
        """
        invocation_arguments = self.get_process_invocation()
        self.process = subprocess.Popen(invocation_arguments)

    def wait(self, timeout=None):
        """Wait for the app to complete then return the return code

        Waits (blocking) for the process to complete. Then returns the return code of the underlying process. If timeout
        is non-None then the process will be killed after waiting for the timeout and another wait of timeout will be
        allowed for the killed process to exit.

        Return:
            return code of the underlying process
        """
        try:
            _, _ = self.process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            self.process.kill()
            _, _ = self.process.wait(timeout=timeout)
        return self.process.returncode

    @abstractmethod
    def get_process_invocation(self) -> List[str]:
        """Run the start-up function

        Run the start-up function unconstrained by the limitations of running in a dedicated subprocess.

        """
        raise NotImplementedError()

    @classmethod
    @gds_plugin_specification
    def register_gds_app_plugin(cls) -> Type["GdsApp"]:
        """Register a gds start-up application

        Plugin hook for registering a plugin that supplies start-up functionality. This functionality will run on start-up
        of the GDS network isolated into a dedicated process.

        Note: users should return the class, not an instance of the class. Needed arguments for instantiation are
        determined from class methods, solicited via the command line, and provided at construction time to the chosen
        instantiation.

        Returns:
            GdsApp subclass
        """
        raise NotImplementedError()


class GdsStandardApp(GdsApp):
    """Standard GDS application that is built upon the StandardPipeline

    Use this class to help build a GdsApp plugin that has a known invocation and starts up the standard pipeline to
    enable standard GDS processes.

    Developers should implement a concrete subclass with the `start(pipeline)` function to run the application with the
    supplied pipeline. The subclass must supply **kwargs parent class constructor and extend a GdsApp plugin:

    ```
    @gds_plugin(GdsApp)
    class MyStandardApp(GdsStandardApp):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
        ...
    ```

    If the plugin requires more arguments beyond the standard pipeline arguments, supply those additional arguments via
    the `get_additional_arguments` method.
    """

    def __init__(self, **kwargs):
        """Take all arguments and store them"""
        super().__init__(**kwargs)

    @classmethod
    def get_additional_arguments(cls) -> Dict[Tuple, Dict[str, str]]:
        """Function to provide additional command line arguments beyond the standard pipeline

        Override this function to provide additional arguments. The form of the arguments are the same as returned by
        standard plugins: a dictionary of tuple flags to argparse kwargs inputs.

        Return:
            dictionary of flag tuple to argparse kwargs
        """
        return {}

    @classmethod
    def init(cls):
        """Allows standard application plugins to initialize before argument parsing is performed"""
        pass

    @final
    @classmethod
    def get_arguments(cls):
        """Get the arguments for this plugin

        This will return the combined arguments needed for the standard pipeline, and those returned from
        `get_additional_arguments()`.
        """
        return {
            **cls.get_additional_arguments(),
            **StandardPipelineParser().get_arguments(),
        }

    @classmethod
    def get_cli_parser(cls):
        """Helper to get a parser for this applications' additional arguments"""
        return BareArgumentParser(
            cls.get_additional_arguments(), getattr(cls, "check_arguments", None)
        )

    @abstractmethod
    def start(self, pipeline: StandardPipeline):
        """Start function to contain behavior based in standard pipeline"""
        raise NotImplementedError()

    def get_process_invocation(self):
        """Return the process invocation for this class' main

        The process invocation of this application is to run cls.main and supply it a reproduced version of the
        arguments needed for the given parsers.  When main is loaded, it will dispatch to the sub-classing plugin's
        start method. The subclassing plugin will already have had the arguments supplied via the PluginParser's
        construction of plugin objects.

        Returns:
            list of arguments to pass to subprocess
        """
        cls = self.__class__.__name__
        module = self.__class__.__module__

        composite_parser = CompositeParser(
            [self.get_cli_parser(), StandardPipelineParser]
        )
        namespace, _, _ = ParserBase.parse_known_args([composite_parser])
        args = composite_parser.reproduce_cli_args(namespace)
        return [sys.executable, "-c", f"import {module}\n{module}.{cls}.main()"] + args

    @classmethod
    def main(cls):
        """Main function used as a generic entrypoint for GdsStandardApp derived GdsApp plugins"""
        try:
            cls.init()
            try:
                Plugins.system(
                    []
                )  # Disable plugin system unless specified through init
            # In the case where `init` sets up the plugin system, we want to pass the assertion
            # triggered by the code above that turns it off in the not-setup case. 
            except AssertionError:
                pass
            plugin_name = getattr(cls, "get_name", lambda: cls.__name__)()
            parsed_arguments, _ = ParserBase.parse_args(
                [cls.get_cli_parser(), StandardPipelineParser, PluginArgumentParser],
                f"{plugin_name}: a standard app plugin",
            )
            pipeline = StandardPipeline()
            # Turn off history and filing
            pipeline.histories.implementation = None
            pipeline.filing = None
            pipeline = StandardPipelineParser.pipeline_factory(
                parsed_arguments, pipeline
            )
            application = cls(
                **cls.get_cli_parser().extract_arguments(parsed_arguments)
            )
            application.start(pipeline)
            sys.exit(0)
        except Exception as e:
            print(f"[ERROR] Error launching {cls.__name__}: {e}", file=sys.stderr)
            raise
            sys.exit(148)


@gds_plugin(GdsApp)
class CustomDataHandlers(GdsStandardApp):
    """Run an app that registers all custom data handlers

    A GdsApp plugin, built using the GdsStandardApp helper, that uses the provided standard pipeline to register each
    custom DataHandler plugin as a consumer of the appropriate type.
    """

    def __init__(self, **kwargs):
        """Required __init__ implementation"""
        super().__init__(**kwargs)

    @classmethod
    def init(cls):
        """Set up the system to use only data_handler plugins"""
        Plugins.system(["data_handler"])

    def start(self, pipeline: StandardPipeline):
        """Iterates over each data handler, registering to the producing decoder"""
        DESCRIPTOR_TO_FUNCTION = {
            "FW_PACKET_TELEM": pipeline.coders.register_channel_consumer,
            "FW_PACKET_LOG": pipeline.coders.register_event_consumer,
            "FW_PACKET_FILE": pipeline.coders.register_file_consumer,
            "FW_PACKET_PACKETIZED_TLM": pipeline.coders.register_packet_consumer,
        }

        data_handlers = Plugins.system().get_feature_classes("data_handler")
        for data_handler_class in data_handlers:
            data_handler = data_handler_class()
            descriptors = data_handler.get_handled_descriptors()
            for descriptor in descriptors:
                DESCRIPTOR_TO_FUNCTION.get(descriptor, lambda discard: discard)(
                    data_handler
                )

    @classmethod
    def get_name(cls):
        """Return the name of this application"""
        return "custom-data-handlers-app"



"""
@brief Class to store a specific packet and associated readings

@date Created July 12, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.data_types.sys_data import SysData


class PktData(SysData):
    """Stores the data from a specific packet receive"""

    def __init__(self, pkt_chs, pkt_time, pkt_temp):
        """
        Constructor.

        Args:
            pkt_chs:  A list of ChData objects. The ChData objects and their
                     order should match the channel order in the pkt_temp
                     object. Each ChData object has the reading for that channel
                     in the packet.
            pkt_time: The time the packet was received (as a TimeType object)
            pkt_temp: PktTemplate object that describes this packet

        Returns:
            An initialized PktData object
        """
        super().__init__()
        self.id = pkt_temp.get_id()
        self.chs = pkt_chs
        self.time = pkt_time
        self.template = pkt_temp

        # Set the packet of all the channels
        for ch in self.chs:
            ch.set_pkt(self)

    def get_chs(self):
        return self.chs

    def get_time(self):
        return self.time

    def get_template(self):
        return self.template

    @staticmethod
    def get_csv_header(verbose=False):
        """
        Get the header for a csv file containing packet data

        Args:
            verbose: (boolean, default=False) Indicates if header should be for
                                              regular or verbose output

        Returns:
            String version of the channel data
        """
        # For csv output, all channels are just printed without regards to
        # packet information
        return ChData.get_csv_header(verbose)

    def get_str(self, time_zone=None, verbose=False, csv=False):
        """
        Convert the packet data to a string

        Args:
            time_zone: (tzinfo, default=None) Timezone to print time in. If
                      time_zone=None, use local time.
            verbose: (boolean, default=False) Prints extra fields if True
            csv: (boolean, default=False) Prints each field with commas between
                                          if true

        Returns:
            String version of the packet data
        """
        pkt_str = ""

        if not csv and verbose:
            pkt_str += "%s: %s (%d) %s{\n" % (
                self.time.to_readable(time_zone),
                self.template.get_name(),
                self.template.get_id(),
                str(self.time),
            )
        elif not csv and not verbose:
            pkt_str += f"{self.time.to_readable(time_zone)}: {self.template.get_name()} {{\n"

        for i, ch in enumerate(self.chs):

            if not csv:
                pkt_str += "\t"

            pkt_str += ch.get_str(time_zone, verbose, csv)

            # Only print newline if we have not just printed the last line
            if i < (len(self.chs) - 1):
                pkt_str += "\n"

        if not csv:
            pkt_str += "\n}"

        return pkt_str

    def __str__(self):
        """
        Convert the pkt data to a human readable string

        Returns:
            String version of the packet data
        """
        return self.get_str()



""" Provides communication adapters for communication with the spacecraft

Adapters are used to adapt the GDS communication layer to a specific communication format (driver). This package
provides two built in adapters: UART and IP used to communicate with a serial device and using IP stack respectively.

If a user wishes to use a different communication format, subclass and implement abstract functions defined by the
BaseAdapter class found in the base module.
"""
